from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="g2",
    description="G2 Buyer Intent API",
    tools={
        "get_intent_signals": Permission.READ,
        "get_product_reviews": Permission.READ,
        "get_comparison_data": Permission.READ,
        "get_category_leaders": Permission.READ,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="jira",
    description="Jira Software REST API",
    tools={
        "list_projects": Permission.READ,
        "get_project": Permission.READ,
        "search_issues": Permission.READ,
        "get_issue": Permission.READ,
        "create_issue": Permission.WRITE,
        "update_issue": Permission.WRITE,
        "transition_issue": Permission.WRITE,
        "get_transitions": Permission.READ,
        "add_comment": Permission.WRITE,
        "get_sprint_data": Permission.READ,
        "get_project_metrics": Permission.READ,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="wordpress",
    description="WordPress REST API",
    tools={
        "create_post": Permission.WRITE,
        "update_post": Permission.WRITE,
        "list_posts": Permission.READ,
        "get_post": Permission.READ,
        "upload_media": Permission.WRITE,
        "list_categories": Permission.READ,
        "create_page": Permission.WRITE,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="s3",
    description="Amazon S3 API",
    tools={
        "upload_document": Permission.WRITE,
        "download_document": Permission.READ,
        "list_objects": Permission.READ,
        "generate_signed_url": Permission.READ,
        "delete_object": Permission.DELETE,
        "copy_object": Permission.WRITE,
    },
)



from __future__ import annotations

from .._http import HttpClient
from .._types import CheckoutResponse, CreateCheckoutParams, CreatePortalParams, PortalResponse, SubscriptionStatus


class BillingClient:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def get_subscription(self) -> SubscriptionStatus:
        """Return the current subscription status for the authenticated developer."""
        data = self._http.get("/v1/billing/subscription")
        return SubscriptionStatus.from_dict(data)

    def create_checkout(self, params: CreateCheckoutParams) -> CheckoutResponse:
        """Create a Stripe Checkout session and return the redirect URL."""
        data = self._http.post("/v1/billing/checkout", params.to_dict())
        return CheckoutResponse.from_dict(data)

    def create_portal(self, params: CreatePortalParams) -> PortalResponse:
        """Create a Stripe Billing Portal session and return the redirect URL."""
        data = self._http.post("/v1/billing/portal", params.to_dict())
        return PortalResponse.from_dict(data)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="epfo",
    description="EPFO (Employees Provident Fund Organisation) API",
    tools={
        "file_ecr": Permission.WRITE,
        "get_uan": Permission.READ,
        "check_claim_status": Permission.READ,
        "download_passbook": Permission.READ,
        "generate_trrn": Permission.WRITE,
        "verify_member": Permission.READ,
    },
)



from __future__ import annotations

import os
from typing import Any, Callable

import httpx

from ._http import HttpClient
from ._types import (
    AuthorizationRequest,
    AuthorizeParams,
    RateLimit,
    RotateKeyResponse,
    SignupParams,
    SignupResponse,
    UpdateDeveloperSettingsParams,
    UpdateDeveloperSettingsResponse,
)
from .resources._agents import AgentsClient
from .resources._audit import AuditClient
from .resources._anomalies import AnomaliesClient
from .resources._scim import ScimClient
from .resources._sso import SsoClient
from .resources._compliance import ComplianceClient
from .resources._grants import GrantsClient
from .resources._tokens import TokensClient
from .resources._webhooks import WebhooksClient
from .resources._billing import BillingClient
from .resources._policies import PoliciesClient
from .resources._principal_sessions import PrincipalSessionsClient
from .resources._vault import VaultClient
from .resources._budgets import BudgetsClient
from .resources._events import EventsClient
from .resources._usage import UsageClient
from .resources._domains import DomainsClient
from .resources._webauthn import WebAuthnClient
from .resources._credentials import CredentialsClient
from .manifest import ToolManifest, Permission, EnforceResult
from ._verify import verify_grant_token
from ._types import VerifyGrantTokenOptions

_DEFAULT_BASE_URL = "https://api.grantex.dev"


class Grantex:
    """Main entry point for the Grantex SDK."""

    agents: AgentsClient
    grants: GrantsClient
    tokens: TokensClient
    audit: AuditClient
    webhooks: WebhooksClient
    billing: BillingClient
    policies: PoliciesClient
    compliance: ComplianceClient
    anomalies: AnomaliesClient
    scim: ScimClient
    sso: SsoClient
    principal_sessions: PrincipalSessionsClient
    vault: VaultClient
    budgets: BudgetsClient
    events: EventsClient
    usage: UsageClient
    domains: DomainsClient
    webauthn: WebAuthnClient
    credentials: CredentialsClient

    @property
    def last_rate_limit(self) -> RateLimit | None:
        return self._http.last_rate_limit

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str = _DEFAULT_BASE_URL,
        timeout: float = 30.0,
        enforce_mode: str = "strict",
    ) -> None:
        resolved_key = api_key or os.environ.get("GRANTEX_API_KEY", "")
        if not resolved_key:
            raise ValueError(
                "Grantex API key is required. Pass `api_key` or set the "
                "GRANTEX_API_KEY environment variable."
            )

        self._enforce_mode = enforce_mode

        self._http = HttpClient(
            base_url=base_url,
            api_key=resolved_key,
            timeout=timeout,
        )

        self.agents = AgentsClient(self._http)
        self.grants = GrantsClient(self._http)
        self.tokens = TokensClient(self._http)
        self.audit = AuditClient(self._http)
        self.webhooks = WebhooksClient(self._http)
        self.billing = BillingClient(self._http)
        self.policies = PoliciesClient(self._http)
        self.compliance = ComplianceClient(self._http)
        self.anomalies = AnomaliesClient(self._http)
        self.scim = ScimClient(self._http)
        self.sso = SsoClient(self._http)
        self.principal_sessions = PrincipalSessionsClient(self._http)
        self.vault = VaultClient(self._http, base_url)
        self.budgets = BudgetsClient(self._http)
        self.events = EventsClient(base_url, resolved_key)
        self.usage = UsageClient(self._http)
        self.domains = DomainsClient(self._http)
        self.webauthn = WebAuthnClient(self._http)
        self.credentials = CredentialsClient(self._http)
        self._manifests: dict[str, ToolManifest] = {}
        self._jwks_uri = f"{base_url.rstrip('/')}/.well-known/jwks.json"

    @staticmethod
    def signup(
        params: SignupParams,
        *,
        base_url: str = _DEFAULT_BASE_URL,
    ) -> SignupResponse:
        """Create a new developer account without an API key.

        Returns the developer ID and a one-time API key.
        """
        url = f"{base_url.rstrip('/')}/v1/signup"
        response = httpx.post(
            url,
            json=params.to_dict(),
            headers={"Accept": "application/json"},
        )
        if not response.is_success:
            body = None
            try:
                body = response.json()
            except Exception:
                pass
            message = (
                body["message"]
                if isinstance(body, dict) and isinstance(body.get("message"), str)
                else f"HTTP {response.status_code}"
            )
            raise ValueError(message)
        return SignupResponse.from_dict(response.json())

    def rotate_key(self) -> RotateKeyResponse:
        """Rotate the current API key. Returns a new key; the old key is invalidated."""
        data = self._http.post("/v1/keys/rotate")
        return RotateKeyResponse.from_dict(data)

    def update_settings(
        self, params: UpdateDeveloperSettingsParams
    ) -> UpdateDeveloperSettingsResponse:
        """Update developer settings (e.g. FIDO/WebAuthn requirements)."""
        data = self._http.patch("/v1/me", params.to_dict())
        return UpdateDeveloperSettingsResponse.from_dict(data)

    def authorize(self, params: AuthorizeParams) -> AuthorizationRequest:
        """Initiate the delegated authorization flow for a user.

        `user_id` is transparently mapped to `principalId` in the request body.
        """
        data = self._http.post("/v1/authorize", params.to_dict())
        return AuthorizationRequest.from_dict(data)

    def load_manifest(self, manifest: ToolManifest) -> None:
        """Load a tool manifest for scope enforcement."""
        self._manifests[manifest.connector] = manifest

    def load_manifests(self, manifests: list[ToolManifest]) -> None:
        """Load multiple tool manifests at once."""
        for m in manifests:
            self._manifests[m.connector] = m

    def enforce(
        self,
        grant_token: str,
        connector: str,
        tool: str,
        amount: float | None = None,
    ) -> EnforceResult:
        """Enforce scope for a tool call.

        1. Verifies the grant token JWT offline (JWKS cached, <1ms after first call)
        2. Looks up the tool's required permission from loaded manifests
        3. Checks if the granted scope level covers the required permission

        Fails closed: unknown connectors/tools are denied by default.

        Example::

            result = grantex.enforce(
                grant_token=token,
                connector="salesforce",
                tool="delete_contact",
            )
            if not result.allowed:
                raise PermissionError(result.reason)
        """
        grant_id = ""
        agent_did = ""
        scopes: list[str] = []
        permission = ""

        # 1. Verify the token offline via JWKS
        try:
            grant = verify_grant_token(
                grant_token,
                VerifyGrantTokenOptions(jwks_uri=self._jwks_uri),
            )
        except Exception as e:
            return self._apply_enforce_mode(EnforceResult(
                allowed=False, reason=f"Token verification failed: {e}",
                grant_id=grant_id, agent_did=agent_did, scopes=scopes,
                permission=permission, connector=connector, tool=tool,
            ))

        grant_id = getattr(grant, "grant_id", "")
        agent_did = getattr(grant, "agent_did", "")
        scopes = list(getattr(grant, "scopes", []))

        def _denied(reason: str) -> EnforceResult:
            return EnforceResult(
                allowed=False, reason=reason,
                grant_id=grant_id, agent_did=agent_did, scopes=scopes,
                permission=permission, connector=connector, tool=tool,
            )

        # 2. Look up manifest for the connector
        manifest = self._manifests.get(connector)
        if not manifest:
            return self._apply_enforce_mode(_denied(f"No manifest loaded for connector '{connector}'. Load a manifest first."))

        # 3. Look up tool permission from manifest
        required_permission = manifest.get_permission(tool)
        if not required_permission:
            return self._apply_enforce_mode(_denied(f"Unknown tool '{tool}' on connector '{connector}'. Tool not found in manifest."))
        permission = required_permission

        # 4. Find the best matching scope for this connector
        granted_permission = self._resolve_granted_permission(scopes, connector)
        if not granted_permission:
            return self._apply_enforce_mode(_denied(f"No scope grants access to connector '{connector}'."))

        # 5. Check permission hierarchy
        if not Permission.covers(granted_permission, required_permission):
            return self._apply_enforce_mode(_denied(f"{granted_permission} scope does not permit {required_permission} operations on {connector}."))

        # 6. Check capped amount if provided
        if amount is not None:
            cap = self._extract_cap(scopes, connector)
            if cap is not None and amount > cap:
                return self._apply_enforce_mode(_denied(f"Amount {amount} exceeds budget cap of {cap} on {connector}."))

        return EnforceResult(
            allowed=True, reason="",
            grant_id=grant_id, agent_did=agent_did, scopes=scopes,
            permission=permission, connector=connector, tool=tool,
        )

    @staticmethod
    def _resolve_granted_permission(
        scopes: list[str], connector: str
    ) -> str | None:
        """Resolve the highest granted permission level for a connector."""
        levels = {"read": 0, "write": 1, "delete": 2, "admin": 3}
        best: str | None = None
        best_level = -1

        for scope in scopes:
            parts = scope.split(":")
            if len(parts) >= 3 and parts[0] in ("tool", "agenticorg") and parts[1] == connector:
                level = levels.get(parts[2], -1)
                if level > best_level:
                    best_level = level
                    best = parts[2]

        return best

    @staticmethod
    def _extract_cap(scopes: list[str], connector: str) -> float | None:
        """Extract budget cap from capped scopes."""
        for scope in scopes:
            parts = scope.split(":")
            if parts[0] == "tool" and len(parts) > 1 and parts[1] == connector:
                try:
                    idx = parts.index("capped")
                    if idx + 1 < len(parts):
                        return float(parts[idx + 1])
                except ValueError:
                    continue
        return None

    def _apply_enforce_mode(self, result: EnforceResult) -> EnforceResult:
        """In permissive mode, allow denied results with a warning."""
        if not result.allowed and self._enforce_mode == "permissive":
            import warnings
            warnings.warn(
                f"[grantex] PERMISSIVE MODE — would deny: {result.reason} "
                f"(connector={result.connector}, tool={result.tool})",
                stacklevel=2,
            )
            return EnforceResult(
                allowed=True,
                reason=result.reason,
                grant_id=result.grant_id,
                agent_did=result.agent_did,
                scopes=result.scopes,
                permission=result.permission,
                connector=result.connector,
                tool=result.tool,
            )
        return result

    def wrap_tool(
        self,
        tool: Any,
        *,
        connector: str,
        tool_name: str,
        grant_token: str | Callable[[], str],
    ) -> Any:
        """Wrap a LangChain StructuredTool with automatic Grantex scope enforcement.

        Before each invocation, the grant token is verified and scopes are checked.
        If denied, raises PermissionError instead of calling the tool.

        Args:
            tool: A LangChain StructuredTool or compatible object with _run/_arun methods.
            connector: Connector name for scope lookup.
            tool_name: Tool name for manifest permission lookup.
            grant_token: Static token string or callable that returns the current token.

        Example::

            protected = grantex.wrap_tool(
                my_tool,
                connector="salesforce",
                tool_name="create_lead",
                grant_token=lambda: state["grant_token"],
            )
        """
        from typing import Callable as _Callable
        grantex = self

        original_run = getattr(tool, '_run', None)
        original_arun = getattr(tool, '_arun', None)

        def _get_token() -> str:
            return grant_token() if callable(grant_token) else grant_token

        def _check() -> None:
            result = grantex.enforce(
                grant_token=_get_token(),
                connector=connector,
                tool=tool_name,
            )
            if not result.allowed:
                raise PermissionError(f"Grantex scope denied: {result.reason}")

        if original_run:
            original = original_run
            def wrapped_run(*args: Any, **kwargs: Any) -> Any:
                _check()
                return original(*args, **kwargs)
            tool._run = wrapped_run

        if original_arun:
            original_async = original_arun
            async def wrapped_arun(*args: Any, **kwargs: Any) -> Any:
                _check()
                return await original_async(*args, **kwargs)
            tool._arun = wrapped_arun

        return tool

    def close(self) -> None:
        self._http.close()

    def __enter__(self) -> Grantex:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="mailchimp",
    description="Mailchimp Email Marketing API",
    tools={
        "list_campaigns": Permission.READ,
        "create_campaign": Permission.WRITE,
        "send_campaign": Permission.WRITE,
        "get_campaign_report": Permission.READ,
        "add_list_member": Permission.WRITE,
        "search_members": Permission.READ,
        "create_template": Permission.WRITE,
        "create_ab_campaign": Permission.WRITE,
        "get_ab_results": Permission.READ,
        "send_winner": Permission.WRITE,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="greenhouse",
    description="Greenhouse Recruiting API",
    tools={
        "list_jobs": Permission.READ,
        "get_candidate": Permission.READ,
        "list_applications": Permission.READ,
        "schedule_interview": Permission.WRITE,
        "create_candidate": Permission.WRITE,
        "advance_application": Permission.WRITE,
        "reject_application": Permission.DELETE,
        "get_scorecards": Permission.READ,
    },
)



"""FastAPI dependency for Grantex scope enforcement.

Usage::

    from grantex import Grantex
    from grantex.fastapi import GrantexEnforcer

    grantex = Grantex(api_key=os.environ["GRANTEX_API_KEY"])
    enforcer = GrantexEnforcer(grantex)

    @app.post("/api/tools/{connector}/{tool}")
    async def execute_tool(
        connector: str,
        tool: str,
        auth: EnforceResult = Depends(enforcer),
    ):
        # auth.allowed is guaranteed True here (raises 403 otherwise)
        ...
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._client import Grantex

from .manifest import EnforceResult


class GrantexEnforcer:
    """FastAPI dependency that enforces Grantex scopes on every request.

    Extracts the Bearer token from the Authorization header, connector
    and tool from path parameters, and calls ``grantex.enforce()``.
    Raises HTTPException(403) if denied.
    """

    def __init__(self, grantex: "Grantex") -> None:
        self._grantex = grantex

    async def __call__(
        self,
        connector: str = "",
        tool: str = "",
        authorization: str = "",
    ) -> EnforceResult:
        """FastAPI dependency callable."""
        # Extract token from Authorization header
        token = ""
        if authorization.startswith("Bearer "):
            token = authorization[7:]

        if not token:
            try:
                from fastapi import HTTPException  # type: ignore[import-not-found,unused-ignore]
                raise HTTPException(status_code=401, detail="Missing grant token")
            except ImportError:
                raise PermissionError("Missing grant token")

        result = self._grantex.enforce(
            grant_token=token,
            connector=connector,
            tool=tool,
        )

        if not result.allowed:
            try:
                from fastapi import HTTPException  # type: ignore[import-not-found,unused-ignore]
                raise HTTPException(
                    status_code=403,
                    detail={"code": "SCOPE_DENIED", "message": result.reason, "connector": connector, "tool": tool},
                )
            except ImportError:
                raise PermissionError(f"Scope denied: {result.reason}")

        return result



"""Tests for SsoClient — legacy and enterprise SSO methods."""
from __future__ import annotations

import pytest
import respx
import httpx

from grantex import Grantex
from grantex._types import (
    CreateSsoConfigParams,
    CreateSsoConnectionParams,
    SsoEnforcementParams,
    SsoLdapCallbackParams,
    SsoOidcCallbackParams,
    SsoSamlCallbackParams,
    UpdateSsoConnectionParams,
)

BASE = "https://api.grantex.dev"

MOCK_SSO_CONFIG = {
    "issuerUrl": "https://idp.example.com",
    "clientId": "client_abc",
    "redirectUri": "https://app.grantex.dev/sso/callback",
    "createdAt": "2026-02-27T00:00:00Z",
    "updatedAt": "2026-02-27T00:00:00Z",
}

MOCK_CONNECTION = {
    "id": "sso_conn_01",
    "developerId": "dev_TEST",
    "name": "Okta OIDC",
    "protocol": "oidc",
    "status": "active",
    "issuerUrl": "https://idp.example.com",
    "clientId": "client_abc",
    "domains": ["example.com"],
    "jitProvisioning": True,
    "enforce": False,
    "groupAttribute": "groups",
    "groupMappings": {"admin": ["admin:*"]},
    "defaultScopes": ["read"],
    "createdAt": "2026-03-01T00:00:00Z",
    "updatedAt": "2026-03-01T00:00:00Z",
}

MOCK_SAML_CONNECTION = {
    "id": "sso_conn_02",
    "developerId": "dev_TEST",
    "name": "Okta SAML",
    "protocol": "saml",
    "status": "active",
    "idpEntityId": "https://idp.example.com/saml",
    "idpSsoUrl": "https://idp.example.com/saml/sso",
    "spEntityId": "https://app.grantex.dev",
    "spAcsUrl": "https://app.grantex.dev/sso/callback/saml",
    "domains": ["corp.com"],
    "jitProvisioning": False,
    "enforce": True,
    "groupMappings": {},
    "defaultScopes": [],
    "createdAt": "2026-03-01T00:00:00Z",
    "updatedAt": "2026-03-01T00:00:00Z",
}

MOCK_SESSION = {
    "id": "sso_sess_01",
    "connectionId": "sso_conn_01",
    "principalId": "princ_01",
    "email": "alice@example.com",
    "name": "Alice Smith",
    "idpSubject": "idp_user_01",
    "groups": ["admin"],
    "mappedScopes": ["admin:*"],
    "expiresAt": "2026-03-02T00:00:00Z",
    "createdAt": "2026-03-01T00:00:00Z",
}

MOCK_CALLBACK_RESULT = {
    "sessionId": "sso_sess_01",
    "email": "alice@corp.com",
    "name": "Alice Smith",
    "sub": "idp_user_01",
    "groups": ["engineering"],
    "mappedScopes": ["read", "write"],
    "principalId": "princ_01",
    "developerId": "dev_TEST",
    "expiresAt": "2026-03-02T00:00:00Z",
}


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


# ── Enterprise SSO Connections ────────────────────────────────────────────


@respx.mock
def test_create_connection(client: Grantex) -> None:
    route = respx.post(f"{BASE}/v1/sso/connections").mock(
        return_value=httpx.Response(201, json=MOCK_CONNECTION)
    )
    params = CreateSsoConnectionParams(
        name="Okta OIDC",
        protocol="oidc",
        issuer_url="https://idp.example.com",
        client_id="client_abc",
        client_secret="secret_xyz",
        domains=["example.com"],
        jit_provisioning=True,
        group_attribute="groups",
        group_mappings={"admin": ["admin:*"]},
        default_scopes=["read"],
    )
    result = client.sso.create_connection(params)

    assert result.id == "sso_conn_01"
    assert result.name == "Okta OIDC"
    assert result.protocol == "oidc"
    assert result.status == "active"
    assert result.issuer_url == "https://idp.example.com"
    assert result.client_id == "client_abc"
    assert result.domains == ("example.com",)
    assert result.jit_provisioning is True
    assert result.enforce is False
    assert result.group_attribute == "groups"
    assert result.group_mappings == {"admin": ["admin:*"]}
    assert result.default_scopes == ("read",)
    body = route.calls[0].request.read()
    assert b"Okta OIDC" in body
    assert b"secret_xyz" in body


@respx.mock
def test_create_saml_connection(client: Grantex) -> None:
    respx.post(f"{BASE}/v1/sso/connections").mock(
        return_value=httpx.Response(201, json=MOCK_SAML_CONNECTION)
    )
    params = CreateSsoConnectionParams(
        name="Okta SAML",
        protocol="saml",
        idp_entity_id="https://idp.example.com/saml",
        idp_sso_url="https://idp.example.com/saml/sso",
        idp_certificate="MIIC...",
        sp_entity_id="https://app.grantex.dev",
        sp_acs_url="https://app.grantex.dev/sso/callback/saml",
        domains=["corp.com"],
    )
    result = client.sso.create_connection(params)

    assert result.id == "sso_conn_02"
    assert result.protocol == "saml"
    assert result.idp_entity_id == "https://idp.example.com/saml"
    assert result.idp_sso_url == "https://idp.example.com/saml/sso"
    assert result.sp_entity_id == "https://app.grantex.dev"
    assert result.sp_acs_url == "https://app.grantex.dev/sso/callback/saml"


@respx.mock
def test_list_connections(client: Grantex) -> None:
    respx.get(f"{BASE}/v1/sso/connections").mock(
        return_value=httpx.Response(200, json={"connections": [MOCK_CONNECTION, MOCK_SAML_CONNECTION]})
    )
    result = client.sso.list_connections()

    assert len(result.connections) == 2
    assert result.connections[0].id == "sso_conn_01"
    assert result.connections[0].protocol == "oidc"
    assert result.connections[1].id == "sso_conn_02"
    assert result.connections[1].protocol == "saml"


@respx.mock
def test_get_connection(client: Grantex) -> None:
    respx.get(f"{BASE}/v1/sso/connections/sso_conn_01").mock(
        return_value=httpx.Response(200, json=MOCK_CONNECTION)
    )
    result = client.sso.get_connection("sso_conn_01")

    assert result.id == "sso_conn_01"
    assert result.developer_id == "dev_TEST"
    assert result.name == "Okta OIDC"
    assert result.created_at == "2026-03-01T00:00:00Z"


@respx.mock
def test_update_connection(client: Grantex) -> None:
    updated = {**MOCK_CONNECTION, "name": "Okta OIDC v2", "status": "testing"}
    route = respx.patch(f"{BASE}/v1/sso/connections/sso_conn_01").mock(
        return_value=httpx.Response(200, json=updated)
    )
    params = UpdateSsoConnectionParams(name="Okta OIDC v2", status="testing")
    result = client.sso.update_connection("sso_conn_01", params)

    assert result.name == "Okta OIDC v2"
    assert result.status == "testing"
    body = route.calls[0].request.read()
    assert b"Okta OIDC v2" in body


@respx.mock
def test_delete_connection(client: Grantex) -> None:
    route = respx.delete(f"{BASE}/v1/sso/connections/sso_conn_01").mock(
        return_value=httpx.Response(204)
    )
    client.sso.delete_connection("sso_conn_01")

    assert route.called


@respx.mock
def test_test_connection(client: Grantex) -> None:
    respx.post(f"{BASE}/v1/sso/connections/sso_conn_01/test").mock(
        return_value=httpx.Response(200, json={
            "success": True,
            "protocol": "oidc",
            "issuer": "https://idp.example.com",
            "authorizationEndpoint": "https://idp.example.com/authorize",
            "tokenEndpoint": "https://idp.example.com/token",
            "jwksUri": "https://idp.example.com/.well-known/jwks.json",
        })
    )
    result = client.sso.test_connection("sso_conn_01")

    assert result.success is True
    assert result.protocol == "oidc"
    assert result.issuer == "https://idp.example.com"
    assert result.authorization_endpoint == "https://idp.example.com/authorize"
    assert result.token_endpoint == "https://idp.example.com/token"
    assert result.jwks_uri == "https://idp.example.com/.well-known/jwks.json"
    assert result.error is None


@respx.mock
def test_test_connection_failure(client: Grantex) -> None:
    respx.post(f"{BASE}/v1/sso/connections/sso_conn_01/test").mock(
        return_value=httpx.Response(200, json={
            "success": False,
            "protocol": "saml",
            "error": "Invalid IdP certificate",
        })
    )
    result = client.sso.test_connection("sso_conn_01")

    assert result.success is False
    assert result.error == "Invalid IdP certificate"


# ── SSO enforcement ───────────────────────────────────────────────────────


@respx.mock
def test_set_enforcement(client: Grantex) -> None:
    route = respx.post(f"{BASE}/v1/sso/enforce").mock(
        return_value=httpx.Response(200, json={"enforce": True, "developerId": "dev_TEST"})
    )
    params = SsoEnforcementParams(enforce=True)
    result = client.sso.set_enforcement(params)

    assert result.enforce is True
    assert result.developer_id == "dev_TEST"
    body = route.calls[0].request.read()
    assert b'"enforce"' in body


@respx.mock
def test_set_enforcement_disable(client: Grantex) -> None:
    respx.post(f"{BASE}/v1/sso/enforce").mock(
        return_value=httpx.Response(200, json={"enforce": False, "developerId": "dev_TEST"})
    )
    result = client.sso.set_enforcement(SsoEnforcementParams(enforce=False))

    assert result.enforce is False


# ── SSO sessions ──────────────────────────────────────────────────────────


@respx.mock
def test_list_sessions(client: Grantex) -> None:
    respx.get(f"{BASE}/v1/sso/sessions").mock(
        return_value=httpx.Response(200, json={"sessions": [MOCK_SESSION]})
    )
    result = client.sso.list_sessions()

    assert len(result.sessions) == 1
    s = result.sessions[0]
    assert s.id == "sso_sess_01"
    assert s.connection_id == "sso_conn_01"
    assert s.principal_id == "princ_01"
    assert s.email == "alice@example.com"
    assert s.name == "Alice Smith"
    assert s.idp_subject == "idp_user_01"
    assert s.groups == ("admin",)
    assert s.mapped_scopes == ("admin:*",)
    assert s.expires_at == "2026-03-02T00:00:00Z"
    assert s.created_at == "2026-03-01T00:00:00Z"


@respx.mock
def test_revoke_session(client: Grantex) -> None:
    route = respx.delete(f"{BASE}/v1/sso/sessions/sso_sess_01").mock(
        return_value=httpx.Response(204)
    )
    client.sso.revoke_session("sso_sess_01")

    assert route.called


# ── SSO login flow ────────────────────────────────────────────────────────


@respx.mock
def test_get_login_url(client: Grantex) -> None:
    route = respx.get(f"{BASE}/sso/login").mock(
        return_value=httpx.Response(200, json={
            "authorizeUrl": "https://idp.example.com/authorize?client_id=abc",
            "protocol": "oidc",
            "connectionId": "sso_conn_01",
        })
    )
    result = client.sso.get_login_url("dev_TEST")

    assert "https://idp.example.com/authorize" in result.authorize_url
    assert result.protocol == "oidc"
    assert result.connection_id == "sso_conn_01"
    url = str(route.calls[0].request.url)
    assert "org=dev_TEST" in url


@respx.mock
def test_get_login_url_with_domain(client: Grantex) -> None:
    route = respx.get(f"{BASE}/sso/login").mock(
        return_value=httpx.Response(200, json={
            "authorizeUrl": "https://idp.example.com/authorize?client_id=abc",
            "protocol": "saml",
            "connectionId": "sso_conn_02",
        })
    )
    result = client.sso.get_login_url("dev_TEST", domain="corp.com")

    assert result.protocol == "saml"
    url = str(route.calls[0].request.url)
    assert "org=dev_TEST" in url
    assert "domain=corp.com" in url


@respx.mock
def test_handle_oidc_callback(client: Grantex) -> None:
    route = respx.post(f"{BASE}/sso/callback/oidc").mock(
        return_value=httpx.Response(200, json=MOCK_CALLBACK_RESULT)
    )
    params = SsoOidcCallbackParams(code="auth_code_xyz", state="state_abc")
    result = client.sso.handle_oidc_callback(params)

    assert result.session_id == "sso_sess_01"
    assert result.email == "alice@corp.com"
    assert result.name == "Alice Smith"
    assert result.sub == "idp_user_01"
    assert result.groups == ("engineering",)
    assert result.mapped_scopes == ("read", "write")
    assert result.principal_id == "princ_01"
    assert result.developer_id == "dev_TEST"
    assert result.expires_at == "2026-03-02T00:00:00Z"
    body = route.calls[0].request.read()
    assert b"auth_code_xyz" in body


@respx.mock
def test_handle_oidc_callback_with_redirect_uri(client: Grantex) -> None:
    route = respx.post(f"{BASE}/sso/callback/oidc").mock(
        return_value=httpx.Response(200, json=MOCK_CALLBACK_RESULT)
    )
    params = SsoOidcCallbackParams(
        code="auth_code_xyz",
        state="state_abc",
        redirect_uri="https://app.example.com/callback",
    )
    result = client.sso.handle_oidc_callback(params)

    assert result.session_id == "sso_sess_01"
    body = route.calls[0].request.read()
    assert b"redirect_uri" in body


@respx.mock
def test_handle_saml_callback(client: Grantex) -> None:
    route = respx.post(f"{BASE}/sso/callback/saml").mock(
        return_value=httpx.Response(200, json=MOCK_CALLBACK_RESULT)
    )
    params = SsoSamlCallbackParams(
        saml_response="PHNhbWxwOlJlc3BvbnNl...",
        relay_state="state_abc",
    )
    result = client.sso.handle_saml_callback(params)

    assert result.session_id == "sso_sess_01"
    assert result.developer_id == "dev_TEST"
    body = route.calls[0].request.read()
    assert b"SAMLResponse" in body
    assert b"RelayState" in body


# ── Legacy methods (backward compatible) ──────────────────────────────────


@respx.mock
def test_create_config(client: Grantex) -> None:
    route = respx.post(f"{BASE}/v1/sso/config").mock(
        return_value=httpx.Response(201, json=MOCK_SSO_CONFIG)
    )
    params = CreateSsoConfigParams(
        issuer_url="https://idp.example.com",
        client_id="client_abc",
        client_secret="secret_xyz",
        redirect_uri="https://app.grantex.dev/sso/callback",
    )
    result = client.sso.create_config(params)

    assert result.issuer_url == "https://idp.example.com"
    assert result.client_id == "client_abc"
    assert result.redirect_uri == "https://app.grantex.dev/sso/callback"
    body = route.calls[0].request.read()
    assert b"client_abc" in body


@respx.mock
def test_get_config(client: Grantex) -> None:
    respx.get(f"{BASE}/v1/sso/config").mock(
        return_value=httpx.Response(200, json=MOCK_SSO_CONFIG)
    )
    result = client.sso.get_config()

    assert result.issuer_url == "https://idp.example.com"
    assert result.created_at == "2026-02-27T00:00:00Z"


@respx.mock
def test_delete_config(client: Grantex) -> None:
    route = respx.delete(f"{BASE}/v1/sso/config").mock(
        return_value=httpx.Response(204)
    )
    client.sso.delete_config()

    assert route.called


@respx.mock
def test_handle_callback(client: Grantex) -> None:
    respx.get(f"{BASE}/sso/callback").mock(
        return_value=httpx.Response(
            200,
            json={
                "email": "alice@corp.com",
                "name": "Alice Smith",
                "sub": "idp_user_01",
                "developerId": "dev_TEST",
            },
        )
    )
    result = client.sso.handle_callback("auth_code_xyz", "state_abc")

    assert result.email == "alice@corp.com"
    assert result.name == "Alice Smith"
    assert result.developer_id == "dev_TEST"


# ── Type serialization ───────────────────────────────────────────────────


def test_create_connection_params_to_dict() -> None:
    params = CreateSsoConnectionParams(
        name="Test",
        protocol="oidc",
        issuer_url="https://idp.example.com",
        client_id="abc",
        client_secret="secret",
    )
    d = params.to_dict()
    assert d["name"] == "Test"
    assert d["protocol"] == "oidc"
    assert d["issuerUrl"] == "https://idp.example.com"
    assert d["clientId"] == "abc"
    assert d["clientSecret"] == "secret"
    assert "idpEntityId" not in d
    assert "domains" not in d


def test_create_connection_params_to_dict_saml() -> None:
    params = CreateSsoConnectionParams(
        name="SAML",
        protocol="saml",
        idp_entity_id="https://idp.example.com/saml",
        idp_sso_url="https://idp.example.com/saml/sso",
        idp_certificate="MIIC...",
        sp_entity_id="https://app.grantex.dev",
        sp_acs_url="https://app.grantex.dev/sso/callback/saml",
        domains=["corp.com"],
        enforce=True,
    )
    d = params.to_dict()
    assert d["protocol"] == "saml"
    assert d["idpEntityId"] == "https://idp.example.com/saml"
    assert d["idpSsoUrl"] == "https://idp.example.com/saml/sso"
    assert d["idpCertificate"] == "MIIC..."
    assert d["spEntityId"] == "https://app.grantex.dev"
    assert d["spAcsUrl"] == "https://app.grantex.dev/sso/callback/saml"
    assert d["domains"] == ["corp.com"]
    assert d["enforce"] is True
    assert "issuerUrl" not in d
    assert "clientId" not in d


def test_update_connection_params_to_dict() -> None:
    params = UpdateSsoConnectionParams(name="Updated", status="inactive")
    d = params.to_dict()
    assert d == {"name": "Updated", "status": "inactive"}


def test_update_connection_params_to_dict_empty() -> None:
    params = UpdateSsoConnectionParams()
    d = params.to_dict()
    assert d == {}


def test_sso_oidc_callback_params_to_dict() -> None:
    params = SsoOidcCallbackParams(code="abc", state="xyz")
    d = params.to_dict()
    assert d == {"code": "abc", "state": "xyz"}
    assert "redirect_uri" not in d


def test_sso_oidc_callback_params_to_dict_with_redirect() -> None:
    params = SsoOidcCallbackParams(code="abc", state="xyz", redirect_uri="https://example.com/cb")
    d = params.to_dict()
    assert d["redirect_uri"] == "https://example.com/cb"


def test_sso_saml_callback_params_to_dict() -> None:
    params = SsoSamlCallbackParams(saml_response="PHNhbWw...", relay_state="state123")
    d = params.to_dict()
    assert d == {"SAMLResponse": "PHNhbWw...", "RelayState": "state123"}


@respx.mock
def test_handle_ldap_callback(client: Grantex) -> None:
    route = respx.post(f"{BASE}/sso/callback/ldap").mock(
        return_value=httpx.Response(200, json=MOCK_CALLBACK_RESULT)
    )
    params = SsoLdapCallbackParams(
        username="carol",
        password="secret",
        connection_id="sso_conn_03",
        org="dev_TEST",
    )
    result = client.sso.handle_ldap_callback(params)

    assert result.session_id == "sso_sess_01"
    assert result.developer_id == "dev_TEST"
    body = route.calls[0].request.read()
    assert b"carol" in body
    assert b"connectionId" in body


@respx.mock
def test_create_ldap_connection(client: Grantex) -> None:
    mock_ldap_connection = {
        **MOCK_CONNECTION,
        "id": "sso_conn_03",
        "name": "Corp LDAP",
        "protocol": "ldap",
        "ldapUrl": "ldap://ldap.corp.com:389",
        "ldapBindDn": "cn=admin,dc=corp,dc=com",
        "ldapSearchBase": "ou=users,dc=corp,dc=com",
        "ldapSearchFilter": "(uid={{username}})",
        "ldapTlsEnabled": True,
    }
    route = respx.post(f"{BASE}/v1/sso/connections").mock(
        return_value=httpx.Response(201, json=mock_ldap_connection)
    )
    params = CreateSsoConnectionParams(
        name="Corp LDAP",
        protocol="ldap",
        ldap_url="ldap://ldap.corp.com:389",
        ldap_bind_dn="cn=admin,dc=corp,dc=com",
        ldap_bind_password="admin_secret",
        ldap_search_base="ou=users,dc=corp,dc=com",
        ldap_search_filter="(uid={{username}})",
        ldap_tls_enabled=True,
        domains=["corp.com"],
    )
    result = client.sso.create_connection(params)

    assert result.id == "sso_conn_03"
    assert result.protocol == "ldap"
    assert result.ldap_url == "ldap://ldap.corp.com:389"
    assert result.ldap_bind_dn == "cn=admin,dc=corp,dc=com"
    assert result.ldap_search_base == "ou=users,dc=corp,dc=com"
    assert result.ldap_search_filter == "(uid={{username}})"
    assert result.ldap_tls_enabled is True
    body = route.calls[0].request.read()
    assert b"ldapUrl" in body
    assert b"ldapBindPassword" in body


def test_sso_ldap_callback_params_to_dict() -> None:
    params = SsoLdapCallbackParams(
        username="carol",
        password="secret",
        connection_id="sso_conn_03",
        org="dev_TEST",
    )
    d = params.to_dict()
    assert d == {
        "username": "carol",
        "password": "secret",
        "connectionId": "sso_conn_03",
        "org": "dev_TEST",
    }


def test_create_connection_params_to_dict_ldap() -> None:
    params = CreateSsoConnectionParams(
        name="LDAP",
        protocol="ldap",
        ldap_url="ldap://ldap.corp.com:389",
        ldap_bind_dn="cn=admin,dc=corp,dc=com",
        ldap_bind_password="admin_secret",
        ldap_search_base="ou=users,dc=corp,dc=com",
        ldap_search_filter="(uid={{username}})",
        ldap_group_search_base="ou=groups,dc=corp,dc=com",
        ldap_group_search_filter="(member={{dn}})",
        ldap_tls_enabled=True,
        domains=["corp.com"],
    )
    d = params.to_dict()
    assert d["protocol"] == "ldap"
    assert d["ldapUrl"] == "ldap://ldap.corp.com:389"
    assert d["ldapBindDn"] == "cn=admin,dc=corp,dc=com"
    assert d["ldapBindPassword"] == "admin_secret"
    assert d["ldapSearchBase"] == "ou=users,dc=corp,dc=com"
    assert d["ldapSearchFilter"] == "(uid={{username}})"
    assert d["ldapGroupSearchBase"] == "ou=groups,dc=corp,dc=com"
    assert d["ldapGroupSearchFilter"] == "(member={{dn}})"
    assert d["ldapTlsEnabled"] is True
    assert d["domains"] == ["corp.com"]
    assert "issuerUrl" not in d
    assert "clientId" not in d
    assert "idpEntityId" not in d


def test_sso_enforcement_params_to_dict() -> None:
    assert SsoEnforcementParams(enforce=True).to_dict() == {"enforce": True}
    assert SsoEnforcementParams(enforce=False).to_dict() == {"enforce": False}



"""Tests for CredentialsClient."""
from __future__ import annotations

import json

import pytest
import respx
import httpx

from grantex import Grantex, ListCredentialsParams, SDJWTPresentParams

MOCK_VC_RECORD = {
    "id": "vc_01",
    "grantId": "grant_01",
    "credentialType": "GrantCredential",
    "format": "vc+sd-jwt",
    "credential": "eyJ0eXAiOiJ2YytzZC1qd3QiLCJhbGciOiJFUzI1NiJ9...",
    "status": "active",
    "issuedAt": "2026-03-01T00:00:00Z",
    "expiresAt": "2026-03-02T00:00:00Z",
}

MOCK_VERIFY_RESULT = {
    "valid": True,
    "credentialType": "GrantCredential",
    "issuer": "did:web:grantex.dev",
    "subject": {"principalId": "user_abc123", "agentDid": "did:grantex:ag_01"},
    "expiresAt": "2026-03-02T00:00:00Z",
    "revoked": False,
}


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


@respx.mock
def test_get_credential(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/v1/credentials/vc_01").mock(
        return_value=httpx.Response(200, json=MOCK_VC_RECORD)
    )
    result = client.credentials.get("vc_01")

    assert result.id == "vc_01"
    assert result.credential_type == "GrantCredential"
    assert result.format == "vc+sd-jwt"


@respx.mock
def test_list_with_params(client: Grantex) -> None:
    respx.get(
        url__regex=r"/v1/credentials\?.*grantId=grant_01"
    ).mock(return_value=httpx.Response(200, json={"credentials": [MOCK_VC_RECORD]}))

    result = client.credentials.list(
        ListCredentialsParams(grant_id="grant_01", status="active")
    )

    assert len(result.credentials) == 1
    assert result.credentials[0].grant_id == "grant_01"


@respx.mock
def test_list_without_params(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/v1/credentials").mock(
        return_value=httpx.Response(200, json={"credentials": []})
    )
    result = client.credentials.list()
    assert len(result.credentials) == 0


@respx.mock
def test_verify_credential(client: Grantex) -> None:
    route = respx.post("https://api.grantex.dev/v1/credentials/verify").mock(
        return_value=httpx.Response(200, json=MOCK_VERIFY_RESULT)
    )
    result = client.credentials.verify("eyJ...")

    assert result.valid is True
    assert result.credential_type == "GrantCredential"
    assert result.issuer == "did:web:grantex.dev"
    assert result.revoked is False

    body = json.loads(route.calls[0].request.content)
    assert body == {"credential": "eyJ..."}


@respx.mock
def test_present_sd_jwt(client: Grantex) -> None:
    mock_result = {
        "valid": True,
        "disclosedClaims": {
            "principalId": "user_abc123",
            "scopes": ["read"],
        },
    }
    route = respx.post("https://api.grantex.dev/v1/credentials/present").mock(
        return_value=httpx.Response(200, json=mock_result)
    )
    result = client.credentials.present(
        SDJWTPresentParams(sd_jwt="eyJ...~disc1~disc2~", nonce="abc123")
    )

    assert result.valid is True
    assert result.disclosed_claims is not None
    assert result.disclosed_claims["principalId"] == "user_abc123"

    body = json.loads(route.calls[0].request.content)
    assert body["sdJwt"] == "eyJ...~disc1~disc2~"
    assert body["nonce"] == "abc123"


@respx.mock
def test_present_sd_jwt_invalid(client: Grantex) -> None:
    mock_result = {
        "valid": False,
        "error": "Invalid SD-JWT signature",
    }
    respx.post("https://api.grantex.dev/v1/credentials/present").mock(
        return_value=httpx.Response(200, json=mock_result)
    )
    result = client.credentials.present(SDJWTPresentParams(sd_jwt="invalid~"))

    assert result.valid is False
    assert result.error == "Invalid SD-JWT signature"



"""Tests for HttpClient edge cases."""
from __future__ import annotations

import pytest
import respx
import httpx

from grantex._http import HttpClient, _parse_rate_limit_headers, _extract_error_code, _extract_error_message


# ── 204 No Content ─────────────────────────────────────────────────────────

@respx.mock
def test_204_returns_none() -> None:
    respx.delete("https://api.grantex.dev/v1/agents/a1").mock(
        return_value=httpx.Response(204)
    )
    client = HttpClient("https://api.grantex.dev", "test-key")
    result = client.delete("/v1/agents/a1")
    assert result is None
    client.close()


# ── Rate limit header parsing ──────────────────────────────────────────────

def test_parse_rate_limit_headers_complete() -> None:
    headers = httpx.Headers({
        "x-ratelimit-limit": "100",
        "x-ratelimit-remaining": "42",
        "x-ratelimit-reset": "1700000000",
        "retry-after": "30",
    })
    rl = _parse_rate_limit_headers(headers)
    assert rl is not None
    assert rl.limit == 100
    assert rl.remaining == 42
    assert rl.reset == 1700000000
    assert rl.retry_after == 30


def test_parse_rate_limit_headers_without_retry_after() -> None:
    headers = httpx.Headers({
        "x-ratelimit-limit": "100",
        "x-ratelimit-remaining": "99",
        "x-ratelimit-reset": "1700000000",
    })
    rl = _parse_rate_limit_headers(headers)
    assert rl is not None
    assert rl.retry_after is None


def test_parse_rate_limit_headers_missing_returns_none() -> None:
    headers = httpx.Headers({"content-type": "application/json"})
    rl = _parse_rate_limit_headers(headers)
    assert rl is None


def test_parse_rate_limit_partial_headers_returns_none() -> None:
    headers = httpx.Headers({"x-ratelimit-limit": "100"})
    rl = _parse_rate_limit_headers(headers)
    assert rl is None


# ── Error extraction helpers ───────────────────────────────────────────────

def test_extract_error_code_from_dict() -> None:
    assert _extract_error_code({"code": "INVALID"}) == "INVALID"


def test_extract_error_code_missing() -> None:
    assert _extract_error_code({"message": "fail"}) is None


def test_extract_error_code_non_dict() -> None:
    assert _extract_error_code("just a string") is None
    assert _extract_error_code(None) is None


def test_extract_error_message_from_message_key() -> None:
    assert _extract_error_message({"message": "Bad request"}, 400) == "Bad request"


def test_extract_error_message_from_error_key() -> None:
    assert _extract_error_message({"error": "Something broke"}, 500) == "Something broke"


def test_extract_error_message_fallback() -> None:
    assert _extract_error_message({}, 502) == "HTTP 502"
    assert _extract_error_message(None, 503) == "HTTP 503"


# ── Custom base_url ────────────────────────────────────────────────────────

@respx.mock
def test_custom_base_url() -> None:
    respx.get("https://custom.example.com/v1/agents").mock(
        return_value=httpx.Response(200, json={"agents": []})
    )
    client = HttpClient("https://custom.example.com", "test-key")
    result = client.get("/v1/agents")
    assert result == {"agents": []}
    client.close()


@respx.mock
def test_base_url_trailing_slash_stripped() -> None:
    respx.get("https://custom.example.com/v1/agents").mock(
        return_value=httpx.Response(200, json={"agents": []})
    )
    client = HttpClient("https://custom.example.com/", "test-key")
    result = client.get("/v1/agents")
    assert result == {"agents": []}
    client.close()


# ── Context manager ────────────────────────────────────────────────────────

@respx.mock
def test_context_manager() -> None:
    respx.get("https://api.grantex.dev/v1/agents").mock(
        return_value=httpx.Response(200, json={"agents": []})
    )
    with HttpClient("https://api.grantex.dev", "test-key") as client:
        result = client.get("/v1/agents")
    assert result == {"agents": []}


# ── Rate limit stored after request ────────────────────────────────────────

@respx.mock
def test_rate_limit_stored_after_successful_request() -> None:
    respx.get("https://api.grantex.dev/v1/agents").mock(
        return_value=httpx.Response(
            200,
            json={"agents": []},
            headers={
                "x-ratelimit-limit": "100",
                "x-ratelimit-remaining": "99",
                "x-ratelimit-reset": "1700000000",
            },
        )
    )
    client = HttpClient("https://api.grantex.dev", "test-key")
    client.get("/v1/agents")
    assert client.last_rate_limit is not None
    assert client.last_rate_limit.remaining == 99
    client.close()


# ── POST with JSON body ───────────────────────────────────────────────────

@respx.mock
def test_post_sends_json_body() -> None:
    import json

    route = respx.post("https://api.grantex.dev/v1/agents").mock(
        return_value=httpx.Response(200, json={"id": "ag_01"})
    )
    client = HttpClient("https://api.grantex.dev", "test-key")
    client.post("/v1/agents", {"name": "test", "scopes": ["read"]})
    body = json.loads(route.calls[0].request.content)
    assert body["name"] == "test"
    assert body["scopes"] == ["read"]
    client.close()


# ── Authorization header sent ──────────────────────────────────────────────

@respx.mock
def test_authorization_header_sent() -> None:
    route = respx.get("https://api.grantex.dev/v1/agents").mock(
        return_value=httpx.Response(200, json={"agents": []})
    )
    client = HttpClient("https://api.grantex.dev", "my-api-key")
    client.get("/v1/agents")
    assert route.calls[0].request.headers["authorization"] == "Bearer my-api-key"
    client.close()



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="docusign",
    description="DocuSign eSignature API",
    tools={
        "send_envelope": Permission.WRITE,
        "get_envelope_status": Permission.READ,
        "void_envelope": Permission.DELETE,
        "download_document": Permission.READ,
        "list_templates": Permission.READ,
        "create_envelope_from_template": Permission.WRITE,
    },
)



"""Tool Manifest & Permission — scope enforcement for AI agent tool calls.

A ToolManifest declares the permission level (read/write/delete/admin)
required for each tool on a connector. The ``enforce()`` method on the
Grantex client uses loaded manifests to check whether a grant token's
scopes allow a given tool call.

Example::

    from grantex import ToolManifest, Permission

    manifest = ToolManifest(
        connector="salesforce",
        tools={
            "query": Permission.READ,
            "create_lead": Permission.WRITE,
            "delete_contact": Permission.DELETE,
        },
    )
"""

from __future__ import annotations

import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Optional


# ── Permission ──────────────────────────────────────────────────────────


class Permission:
    """Permission levels for tool operations.

    Hierarchy: ``admin > delete > write > read``

    A ``write`` scope covers ``read`` + ``write`` tools.
    A ``delete`` scope covers ``read`` + ``write`` + ``delete`` tools.
    An ``admin`` scope covers everything.
    """

    READ = "read"
    WRITE = "write"
    DELETE = "delete"
    ADMIN = "admin"

    _LEVELS: Dict[str, int] = {
        "read": 0,
        "write": 1,
        "delete": 2,
        "admin": 3,
    }

    @staticmethod
    def covers(granted: str, required: str) -> bool:
        """Check whether a granted permission level covers the required level."""
        granted_level = Permission._LEVELS.get(granted, -1)
        required_level = Permission._LEVELS.get(required, 99)
        return granted_level >= required_level

    @staticmethod
    def is_valid(value: str) -> bool:
        """Check whether a string is a valid permission level."""
        return value in Permission._LEVELS


# ── ToolManifest ────────────────────────────────────────────────────────


class ToolManifest:
    """Declares the required permission level for each tool on a connector.

    Load manifests via ``grantex.load_manifest()`` and they will be used
    automatically by ``grantex.enforce()``.
    """

    def __init__(
        self,
        connector: str,
        tools: Dict[str, str],
        version: str = "1.0.0",
        description: str = "",
    ) -> None:
        if not connector:
            raise ValueError("ToolManifest: connector name is required")
        if not tools:
            raise ValueError("ToolManifest: at least one tool is required")
        for name, perm in tools.items():
            if not Permission.is_valid(perm):
                raise ValueError(
                    f'ToolManifest: invalid permission "{perm}" for tool "{name}". '
                    f"Must be one of: {', '.join(Permission._LEVELS.keys())}"
                )

        self.connector = connector
        self.tools: Dict[str, str] = dict(tools)
        self.version = version
        self.description = description

    def get_permission(self, tool_name: str) -> Optional[str]:
        """Get the declared permission for a tool. Returns None if not found."""
        return self.tools.get(tool_name)

    def add_tool(self, tool_name: str, permission: str) -> None:
        """Add or update a tool's permission in this manifest."""
        if not Permission.is_valid(permission):
            raise ValueError(f"Invalid permission: {permission}")
        self.tools[tool_name] = permission

    @property
    def tool_count(self) -> int:
        """Number of tools in this manifest."""
        return len(self.tools)

    @classmethod
    def from_file(cls, path: str) -> "ToolManifest":
        """Load a ToolManifest from a JSON file.

        Expected shape::

            { "connector": "salesforce", "tools": { "query": "read", ... } }
        """
        data = json.loads(Path(path).read_text(encoding="utf-8"))
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ToolManifest":
        """Create a ToolManifest from a dict (e.g., parsed JSON)."""
        connector = data.get("connector")
        tools = data.get("tools")
        if not connector or not tools:
            raise ValueError('ToolManifest: missing "connector" or "tools" field')
        return cls(
            connector=str(connector),
            tools=dict(tools),
            version=str(data.get("version", "1.0.0")),
            description=str(data.get("description", "")),
        )


# ── EnforceResult ───────────────────────────────────────────────────────


@dataclass
class EnforceResult:
    """Result of a ``grantex.enforce()`` call."""

    allowed: bool
    """Whether the tool call is permitted."""

    reason: str
    """Human-readable reason if denied."""

    grant_id: str = ""
    """Grant ID from the JWT."""

    agent_did: str = ""
    """Agent DID from the JWT."""

    scopes: list[str] = field(default_factory=list)
    """All scopes from the grant token."""

    permission: str = ""
    """Resolved permission for the requested tool."""

    connector: str = ""
    """Connector name."""

    tool: str = ""
    """Tool name."""



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="keka",
    description="Keka HR API",
    tools={
        "get_employee": Permission.READ,
        "list_employees": Permission.READ,
        "run_payroll": Permission.ADMIN,
        "get_leave_balance": Permission.READ,
        "post_reimbursement": Permission.WRITE,
        "get_attendance_summary": Permission.READ,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="whatsapp",
    description="WhatsApp Business API",
    tools={
        "send_template_message": Permission.WRITE,
        "send_text_message": Permission.WRITE,
        "send_media_message": Permission.WRITE,
        "get_message_templates": Permission.READ,
        "get_business_profile": Permission.READ,
    },
)



"""Tests for DomainsClient — create, list, verify, delete."""
from __future__ import annotations

import json

import pytest
import respx
import httpx

from grantex import Grantex
from grantex._errors import GrantexApiError
from grantex.resources._domains import CreateDomainParams


BASE_URL = "https://api.grantex.dev"

MOCK_DOMAIN_CREATE = {
    "id": "dom_01",
    "domain": "example.com",
    "verified": False,
    "verificationToken": "grantex-verify-abc123",
    "instructions": "Add a DNS TXT record: _grantex.example.com = grantex-verify-abc123",
}

MOCK_DOMAIN_ENTRY = {
    "id": "dom_01",
    "domain": "example.com",
    "verified": True,
    "verifiedAt": "2024-06-01T12:00:00Z",
    "createdAt": "2024-06-01T10:00:00Z",
}

MOCK_DOMAIN_ENTRY_UNVERIFIED = {
    "id": "dom_02",
    "domain": "staging.example.com",
    "verified": False,
    "verifiedAt": None,
    "createdAt": "2024-06-02T10:00:00Z",
}

MOCK_VERIFY_SUCCESS = {"verified": True, "message": "Domain verified successfully"}
MOCK_VERIFY_FAILURE = {"verified": False, "message": "DNS verification failed. Ensure TXT record is set."}


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


# ── Create Domain ─────────────────────────────────────────────────────────

@respx.mock
def test_create_domain(client: Grantex) -> None:
    route = respx.post(f"{BASE_URL}/v1/domains").mock(
        return_value=httpx.Response(200, json=MOCK_DOMAIN_CREATE)
    )
    result = client.domains.create(CreateDomainParams(domain="example.com"))

    assert result.id == "dom_01"
    assert result.domain == "example.com"
    assert result.verified is False
    assert result.verification_token == "grantex-verify-abc123"
    assert "TXT record" in result.instructions
    assert "example.com" in result.instructions

    body = json.loads(route.calls[0].request.content)
    assert body["domain"] == "example.com"


@respx.mock
def test_create_domain_sends_correct_body(client: Grantex) -> None:
    route = respx.post(f"{BASE_URL}/v1/domains").mock(
        return_value=httpx.Response(200, json=MOCK_DOMAIN_CREATE)
    )
    client.domains.create(CreateDomainParams(domain="api.mysite.io"))

    body = json.loads(route.calls[0].request.content)
    assert body == {"domain": "api.mysite.io"}


@respx.mock
def test_create_domain_plan_limit_error(client: Grantex) -> None:
    respx.post(f"{BASE_URL}/v1/domains").mock(
        return_value=httpx.Response(
            402,
            json={
                "message": "Custom domains require Enterprise plan",
                "code": "PLAN_LIMIT_EXCEEDED",
            },
        )
    )
    with pytest.raises(GrantexApiError) as exc_info:
        client.domains.create(CreateDomainParams(domain="premium.example.com"))
    assert exc_info.value.status_code == 402
    assert exc_info.value.code == "PLAN_LIMIT_EXCEEDED"


@respx.mock
def test_create_domain_conflict_error(client: Grantex) -> None:
    respx.post(f"{BASE_URL}/v1/domains").mock(
        return_value=httpx.Response(
            409,
            json={"message": "Domain already registered", "code": "CONFLICT"},
        )
    )
    with pytest.raises(GrantexApiError) as exc_info:
        client.domains.create(CreateDomainParams(domain="duplicate.example.com"))
    assert exc_info.value.status_code == 409
    assert exc_info.value.code == "CONFLICT"


# ── List Domains ──────────────────────────────────────────────────────────

@respx.mock
def test_list_domains(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/domains").mock(
        return_value=httpx.Response(200, json={"domains": [MOCK_DOMAIN_ENTRY]})
    )
    result = client.domains.list()

    assert len(result.domains) == 1
    assert result.domains[0].id == "dom_01"
    assert result.domains[0].domain == "example.com"
    assert result.domains[0].verified is True
    assert result.domains[0].verified_at == "2024-06-01T12:00:00Z"
    assert result.domains[0].created_at == "2024-06-01T10:00:00Z"


@respx.mock
def test_list_domains_multiple(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/domains").mock(
        return_value=httpx.Response(
            200, json={"domains": [MOCK_DOMAIN_ENTRY, MOCK_DOMAIN_ENTRY_UNVERIFIED]}
        )
    )
    result = client.domains.list()

    assert len(result.domains) == 2
    assert result.domains[0].verified is True
    assert result.domains[1].verified is False
    assert result.domains[1].verified_at is None
    assert result.domains[1].domain == "staging.example.com"


@respx.mock
def test_list_domains_empty(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/domains").mock(
        return_value=httpx.Response(200, json={"domains": []})
    )
    result = client.domains.list()
    assert len(result.domains) == 0


# ── Verify Domain ────────────────────────────────────────────────────────

@respx.mock
def test_verify_domain_success(client: Grantex) -> None:
    respx.post(f"{BASE_URL}/v1/domains/dom_01/verify").mock(
        return_value=httpx.Response(200, json=MOCK_VERIFY_SUCCESS)
    )
    result = client.domains.verify("dom_01")

    assert result.verified is True
    assert result.message == "Domain verified successfully"


@respx.mock
def test_verify_domain_failure(client: Grantex) -> None:
    respx.post(f"{BASE_URL}/v1/domains/dom_01/verify").mock(
        return_value=httpx.Response(
            400,
            json={
                "message": "DNS verification failed. Ensure TXT record is set.",
                "code": "VERIFICATION_FAILED",
            },
        )
    )
    with pytest.raises(GrantexApiError) as exc_info:
        client.domains.verify("dom_01")
    assert exc_info.value.status_code == 400
    assert exc_info.value.code == "VERIFICATION_FAILED"


@respx.mock
def test_verify_domain_already_verified(client: Grantex) -> None:
    respx.post(f"{BASE_URL}/v1/domains/dom_01/verify").mock(
        return_value=httpx.Response(
            200, json={"verified": True, "message": "Domain already verified"}
        )
    )
    result = client.domains.verify("dom_01")
    assert result.verified is True
    assert result.message == "Domain already verified"


@respx.mock
def test_verify_domain_not_found(client: Grantex) -> None:
    respx.post(f"{BASE_URL}/v1/domains/dom_nonexistent/verify").mock(
        return_value=httpx.Response(
            404, json={"message": "Domain not found", "code": "NOT_FOUND"}
        )
    )
    with pytest.raises(GrantexApiError) as exc_info:
        client.domains.verify("dom_nonexistent")
    assert exc_info.value.status_code == 404


@respx.mock
def test_verify_domain_sends_post_to_correct_url(client: Grantex) -> None:
    route = respx.post(f"{BASE_URL}/v1/domains/dom_xyz/verify").mock(
        return_value=httpx.Response(200, json=MOCK_VERIFY_SUCCESS)
    )
    client.domains.verify("dom_xyz")
    assert route.called
    assert str(route.calls[0].request.url) == f"{BASE_URL}/v1/domains/dom_xyz/verify"


# ── Delete Domain ────────────────────────────────────────────────────────

@respx.mock
def test_delete_domain(client: Grantex) -> None:
    route = respx.delete(f"{BASE_URL}/v1/domains/dom_01").mock(
        return_value=httpx.Response(204)
    )
    result = client.domains.delete("dom_01")
    assert result is None
    assert route.called


@respx.mock
def test_delete_domain_not_found(client: Grantex) -> None:
    respx.delete(f"{BASE_URL}/v1/domains/dom_nonexistent").mock(
        return_value=httpx.Response(
            404, json={"message": "Domain not found", "code": "NOT_FOUND"}
        )
    )
    with pytest.raises(GrantexApiError) as exc_info:
        client.domains.delete("dom_nonexistent")
    assert exc_info.value.status_code == 404


@respx.mock
def test_delete_domain_sends_correct_method(client: Grantex) -> None:
    route = respx.delete(f"{BASE_URL}/v1/domains/dom_abc").mock(
        return_value=httpx.Response(204)
    )
    client.domains.delete("dom_abc")
    assert route.calls[0].request.method == "DELETE"



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="netsuite",
    description="NetSuite ERP API",
    tools={
        "create_invoice": Permission.WRITE,
        "get_invoice": Permission.READ,
        "create_journal_entry": Permission.WRITE,
        "get_account_balance": Permission.READ,
        "create_vendor_bill": Permission.WRITE,
        "create_purchase_order": Permission.WRITE,
        "get_trial_balance": Permission.READ,
        "search_records": Permission.READ,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="mca_portal",
    description="MCA (Ministry of Corporate Affairs) Portal API",
    tools={
        "file_annual_return": Permission.WRITE,
        "complete_director_kyc": Permission.WRITE,
        "fetch_company_master_data": Permission.READ,
        "file_charge_satisfaction": Permission.WRITE,
    },
)



from __future__ import annotations

from typing import Any

from .._http import HttpClient
from .._types import (
    CreateScimUserParams,
    ListScimTokensResponse,
    ScimListResponse,
    ScimToken,
    ScimTokenWithSecret,
    ScimUser,
)


class ScimClient:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    # ── SCIM token management ─────────────────────────────────────────────

    def create_token(self, label: str) -> ScimTokenWithSecret:
        """Create a new SCIM bearer token. The raw token is returned once only."""
        data = self._http.post("/v1/scim/tokens", {"label": label})
        return ScimTokenWithSecret.from_dict(data)

    def list_tokens(self) -> ListScimTokensResponse:
        """List all SCIM tokens for this developer org (without raw secrets)."""
        data = self._http.get("/v1/scim/tokens")
        return ListScimTokensResponse.from_dict(data)

    def revoke_token(self, token_id: str) -> None:
        """Revoke a SCIM token by ID."""
        self._http.delete(f"/v1/scim/tokens/{token_id}")

    # ── SCIM 2.0 Users ────────────────────────────────────────────────────

    def list_users(
        self,
        *,
        start_index: int | None = None,
        count: int | None = None,
    ) -> ScimListResponse:
        """List provisioned users (SCIM 2.0 ListResponse)."""
        params: list[str] = []
        if start_index is not None:
            params.append(f"startIndex={start_index}")
        if count is not None:
            params.append(f"count={count}")
        path = "/scim/v2/Users"
        if params:
            path = f"{path}?{'&'.join(params)}"
        data = self._http.get(path)
        return ScimListResponse.from_dict(data)

    def get_user(self, user_id: str) -> ScimUser:
        """Get a single provisioned user by ID."""
        data = self._http.get(f"/scim/v2/Users/{user_id}")
        return ScimUser.from_dict(data)

    def create_user(self, params: CreateScimUserParams) -> ScimUser:
        """Provision a new user."""
        data = self._http.post("/scim/v2/Users", params.to_dict())
        return ScimUser.from_dict(data)

    def replace_user(self, user_id: str, params: CreateScimUserParams) -> ScimUser:
        """Full replace of a user (PUT)."""
        data = self._http.put(f"/scim/v2/Users/{user_id}", params.to_dict())
        return ScimUser.from_dict(data)

    def update_user(
        self,
        user_id: str,
        operations: list[dict[str, Any]],
    ) -> ScimUser:
        """Partial update via SCIM Operations (PATCH)."""
        data = self._http.patch(f"/scim/v2/Users/{user_id}", {"Operations": operations})
        return ScimUser.from_dict(data)

    def delete_user(self, user_id: str) -> None:
        """Deprovision a user (DELETE)."""
        self._http.delete(f"/scim/v2/Users/{user_id}")

    # Keep reference to avoid unused-import
    _ScimToken = ScimToken



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="github",
    description="GitHub REST API",
    tools={
        "list_repos": Permission.READ,
        "get_repo": Permission.READ,
        "list_repository_issues": Permission.READ,
        "create_issue": Permission.WRITE,
        "create_pull_request": Permission.WRITE,
        "get_repository_statistics": Permission.READ,
        "search_code": Permission.READ,
        "create_release": Permission.WRITE,
        "trigger_github_action_workflow": Permission.WRITE,
    },
)



"""Tests for ScimClient."""
from __future__ import annotations

import pytest
import respx
import httpx

from grantex import Grantex
from grantex._types import CreateScimUserParams

MOCK_SCIM_USER = {
    "schemas": ["urn:ietf:params:scim:schemas:core:2.0:User"],
    "id": "scimuser_01",
    "userName": "alice@corp.com",
    "displayName": "Alice Smith",
    "externalId": "ext_abc",
    "active": True,
    "emails": [{"value": "alice@corp.com", "primary": True}],
    "meta": {
        "resourceType": "User",
        "created": "2026-02-27T00:00:00Z",
        "lastModified": "2026-02-27T00:00:00Z",
    },
}

MOCK_TOKEN = {
    "id": "scimtok_01",
    "label": "Okta",
    "token": "raw_secret_abc",
    "createdAt": "2026-02-27T00:00:00Z",
    "lastUsedAt": None,
}


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


# ── SCIM token management ─────────────────────────────────────────────────────

@respx.mock
def test_create_token(client: Grantex) -> None:
    respx.post("https://api.grantex.dev/v1/scim/tokens").mock(
        return_value=httpx.Response(201, json=MOCK_TOKEN)
    )
    result = client.scim.create_token("Okta")

    assert result.id == "scimtok_01"
    assert result.label == "Okta"
    assert result.token == "raw_secret_abc"
    assert result.last_used_at is None


@respx.mock
def test_list_tokens(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/v1/scim/tokens").mock(
        return_value=httpx.Response(200, json={"tokens": [MOCK_TOKEN]})
    )
    result = client.scim.list_tokens()

    assert len(result.tokens) == 1
    assert result.tokens[0].id == "scimtok_01"


@respx.mock
def test_revoke_token(client: Grantex) -> None:
    route = respx.delete("https://api.grantex.dev/v1/scim/tokens/scimtok_01").mock(
        return_value=httpx.Response(204)
    )
    client.scim.revoke_token("scimtok_01")

    assert route.called


# ── SCIM 2.0 Users ────────────────────────────────────────────────────────────

@respx.mock
def test_list_users(client: Grantex) -> None:
    list_resp = {
        "totalResults": 1,
        "startIndex": 1,
        "itemsPerPage": 100,
        "Resources": [MOCK_SCIM_USER],
    }
    respx.get("https://api.grantex.dev/scim/v2/Users").mock(
        return_value=httpx.Response(200, json=list_resp)
    )
    result = client.scim.list_users()

    assert result.total_results == 1
    assert result.resources[0].user_name == "alice@corp.com"
    assert result.resources[0].active is True


@respx.mock
def test_list_users_with_pagination(client: Grantex) -> None:
    route = respx.get("https://api.grantex.dev/scim/v2/Users").mock(
        return_value=httpx.Response(
            200,
            json={"totalResults": 0, "startIndex": 11, "itemsPerPage": 10, "Resources": []},
        )
    )
    client.scim.list_users(start_index=11, count=10)

    url = str(route.calls[0].request.url)
    assert "startIndex=11" in url
    assert "count=10" in url


@respx.mock
def test_get_user(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/scim/v2/Users/scimuser_01").mock(
        return_value=httpx.Response(200, json=MOCK_SCIM_USER)
    )
    result = client.scim.get_user("scimuser_01")

    assert result.id == "scimuser_01"
    assert result.external_id == "ext_abc"
    assert result.display_name == "Alice Smith"
    assert result.meta.resource_type == "User"


@respx.mock
def test_create_user(client: Grantex) -> None:
    route = respx.post("https://api.grantex.dev/scim/v2/Users").mock(
        return_value=httpx.Response(201, json=MOCK_SCIM_USER)
    )
    params = CreateScimUserParams(
        user_name="alice@corp.com",
        display_name="Alice Smith",
        emails=[{"value": "alice@corp.com", "primary": True}],
    )
    result = client.scim.create_user(params)

    assert result.id == "scimuser_01"
    body = route.calls[0].request.read()
    assert b"alice@corp.com" in body


@respx.mock
def test_replace_user(client: Grantex) -> None:
    route = respx.put("https://api.grantex.dev/scim/v2/Users/scimuser_01").mock(
        return_value=httpx.Response(200, json=MOCK_SCIM_USER)
    )
    params = CreateScimUserParams(user_name="alice@corp.com")
    client.scim.replace_user("scimuser_01", params)

    assert route.called


@respx.mock
def test_update_user(client: Grantex) -> None:
    deactivated = {**MOCK_SCIM_USER, "active": False}
    respx.patch("https://api.grantex.dev/scim/v2/Users/scimuser_01").mock(
        return_value=httpx.Response(200, json=deactivated)
    )
    result = client.scim.update_user(
        "scimuser_01",
        [{"op": "replace", "path": "active", "value": False}],
    )

    assert result.active is False


@respx.mock
def test_delete_user(client: Grantex) -> None:
    route = respx.delete("https://api.grantex.dev/scim/v2/Users/scimuser_01").mock(
        return_value=httpx.Response(204)
    )
    client.scim.delete_user("scimuser_01")

    assert route.called



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="brandwatch",
    description="Brandwatch Social Listening API",
    tools={
        "get_mentions": Permission.READ,
        "get_mention_summary": Permission.READ,
        "get_share_of_voice": Permission.READ,
        "create_alert": Permission.WRITE,
        "export_report": Permission.READ,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="salesforce",
    description="Salesforce CRM REST API",
    tools={
        "create_lead": Permission.WRITE,
        "update_opportunity": Permission.WRITE,
        "query": Permission.READ,
        "create_task": Permission.WRITE,
        "get_account": Permission.READ,
        "list_opportunities": Permission.READ,
    },
)



from __future__ import annotations

from typing import Any

from .._http import HttpClient
from .._types import (
    AllocateBudgetParams,
    BudgetAllocation,
    BudgetTransactionsResponse,
    DebitBudgetParams,
    DebitBudgetResponse,
)


class BudgetsClient:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def allocate(self, params: AllocateBudgetParams) -> BudgetAllocation:
        """Allocate a spending budget for a grant."""
        data = self._http.post("/v1/budget/allocate", params.to_dict())
        return BudgetAllocation.from_dict(data)

    def debit(self, params: DebitBudgetParams) -> DebitBudgetResponse:
        """Debit an amount from a grant's budget."""
        data = self._http.post("/v1/budget/debit", params.to_dict())
        return DebitBudgetResponse.from_dict(data)

    def balance(self, grant_id: str) -> BudgetAllocation:
        """Get the current budget balance for a grant."""
        data = self._http.get(f"/v1/budget/balance/{grant_id}")
        return BudgetAllocation.from_dict(data)

    def transactions(
        self,
        grant_id: str,
        page: int = 1,
        page_size: int = 20,
    ) -> BudgetTransactionsResponse:
        """List budget transactions for a grant."""
        params: list[str] = []
        params.append(f"page={page}")
        params.append(f"pageSize={page_size}")
        qs = "&".join(params)
        data = self._http.get(f"/v1/budget/transactions/{grant_id}?{qs}")
        return BudgetTransactionsResponse.from_dict(data)



"""Tests for WebAuthnClient."""
from __future__ import annotations

import json

import pytest
import respx
import httpx

from grantex import Grantex

MOCK_REGISTRATION_OPTIONS = {
    "challengeId": "ch_01",
    "publicKey": {
        "rp": {"name": "Grantex", "id": "grantex.dev"},
        "challenge": "dGVzdC1jaGFsbGVuZ2U",
        "pubKeyCredParams": [{"type": "public-key", "alg": -7}],
    },
}

MOCK_CREDENTIAL = {
    "id": "cred_01",
    "principalId": "user_abc123",
    "deviceName": "YubiKey 5",
    "backedUp": False,
    "transports": ["usb"],
    "createdAt": "2026-03-01T00:00:00Z",
    "lastUsedAt": None,
}


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


@respx.mock
def test_register_options(client: Grantex) -> None:
    route = respx.post("https://api.grantex.dev/v1/webauthn/register/options").mock(
        return_value=httpx.Response(200, json=MOCK_REGISTRATION_OPTIONS)
    )
    result = client.webauthn.register_options(principal_id="user_abc123")

    assert result.challenge_id == "ch_01"
    assert "rp" in result.public_key

    body = json.loads(route.calls[0].request.content)
    assert body["principalId"] == "user_abc123"


@respx.mock
def test_register_verify(client: Grantex) -> None:
    route = respx.post("https://api.grantex.dev/v1/webauthn/register/verify").mock(
        return_value=httpx.Response(200, json=MOCK_CREDENTIAL)
    )
    from grantex import WebAuthnRegistrationVerifyParams

    result = client.webauthn.register_verify(
        WebAuthnRegistrationVerifyParams(
            challenge_id="ch_01",
            response={"id": "rawId", "rawId": "rawId", "type": "public-key"},
            device_name="YubiKey 5",
        )
    )

    assert result.id == "cred_01"
    assert result.principal_id == "user_abc123"
    assert result.device_name == "YubiKey 5"

    body = json.loads(route.calls[0].request.content)
    assert body["challengeId"] == "ch_01"
    assert body["deviceName"] == "YubiKey 5"


@respx.mock
def test_list_credentials(client: Grantex) -> None:
    respx.get(
        "https://api.grantex.dev/v1/webauthn/credentials?principalId=user_abc123"
    ).mock(return_value=httpx.Response(200, json={"credentials": [MOCK_CREDENTIAL]}))

    result = client.webauthn.list_credentials("user_abc123")

    assert len(result.credentials) == 1
    assert result.credentials[0].device_name == "YubiKey 5"


@respx.mock
def test_delete_credential(client: Grantex) -> None:
    respx.delete("https://api.grantex.dev/v1/webauthn/credentials/cred_01").mock(
        return_value=httpx.Response(204)
    )
    result = client.webauthn.delete_credential("cred_01")
    assert result is None



"""Tests for Grantex.enforce() scope enforcement."""
from __future__ import annotations

from unittest.mock import patch

import pytest

from grantex import Grantex, ToolManifest, Permission, EnforceResult
from grantex._errors import GrantexTokenError
from grantex._types import VerifiedGrant


# ── Helpers ──────────────────────────────────────────────────────────────────


def _make_verified_grant(
    scopes: tuple[str, ...] = ("tool:salesforce:write",),
    grant_id: str = "grnt_01HXYZ",
    agent_did: str = "did:grantex:ag_01HXYZ123abc",
) -> VerifiedGrant:
    """Create a VerifiedGrant with the given scopes."""
    return VerifiedGrant(
        token_id="tok_01HXYZ987xyz",
        grant_id=grant_id,
        principal_id="user_abc123",
        agent_did=agent_did,
        developer_id="org_test",
        scopes=scopes,
        issued_at=1709000000,
        expires_at=9999999999,
    )


def _salesforce_manifest() -> ToolManifest:
    return ToolManifest(
        connector="salesforce",
        tools={
            "query": Permission.READ,
            "create_lead": Permission.WRITE,
            "delete_contact": Permission.DELETE,
            "manage_org": Permission.ADMIN,
        },
    )


def _github_manifest() -> ToolManifest:
    return ToolManifest(
        connector="github",
        tools={
            "list_repos": Permission.READ,
            "create_issue": Permission.WRITE,
            "delete_repo": Permission.DELETE,
        },
    )


# ── Fixtures ─────────────────────────────────────────────────────────────────


@pytest.fixture()
def client() -> Grantex:
    c = Grantex(api_key="test-key")
    c.load_manifest(_salesforce_manifest())
    return c


# ── Basic allow/deny ────────────────────────────────────────────────────────


class TestEnforceBasic:
    @patch("grantex._client.verify_grant_token")
    def test_allowed_when_scope_matches(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead")

        assert result.allowed is True
        assert result.reason == ""
        assert result.connector == "salesforce"
        assert result.tool == "create_lead"
        assert result.permission == "write"
        assert result.grant_id == "grnt_01HXYZ"
        assert result.agent_did == "did:grantex:ag_01HXYZ123abc"

    @patch("grantex._client.verify_grant_token")
    def test_denied_when_scope_insufficient(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:read",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead")

        assert result.allowed is False
        assert "does not permit" in result.reason
        assert result.permission == "write"

    @patch("grantex._client.verify_grant_token")
    def test_denied_for_unknown_connector(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        result = client.enforce("fake.jwt.token", "unknown_connector", "some_tool")

        assert result.allowed is False
        assert "No manifest loaded" in result.reason
        assert result.connector == "unknown_connector"

    @patch("grantex._client.verify_grant_token")
    def test_denied_for_unknown_tool(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "nonexistent_tool")

        assert result.allowed is False
        assert "Unknown tool" in result.reason
        assert result.tool == "nonexistent_tool"


# ── Permission hierarchy ────────────────────────────────────────────────────


class TestEnforcePermissionHierarchy:
    @patch("grantex._client.verify_grant_token")
    def test_write_scope_allows_read_tool(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "query")

        assert result.allowed is True
        assert result.permission == "read"

    @patch("grantex._client.verify_grant_token")
    def test_read_scope_denies_write_tool(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:read",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead")

        assert result.allowed is False
        assert "does not permit" in result.reason

    @patch("grantex._client.verify_grant_token")
    def test_delete_scope_allows_write_tool(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:delete",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead")

        assert result.allowed is True

    @patch("grantex._client.verify_grant_token")
    def test_delete_scope_allows_read_tool(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:delete",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "query")

        assert result.allowed is True

    @patch("grantex._client.verify_grant_token")
    def test_admin_scope_allows_all_tools(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:admin",)
        )
        for tool in ("query", "create_lead", "delete_contact", "manage_org"):
            result = client.enforce("fake.jwt.token", "salesforce", tool)
            assert result.allowed is True, f"admin should allow {tool}"

    @patch("grantex._client.verify_grant_token")
    def test_read_scope_denies_delete_tool(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:read",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "delete_contact")

        assert result.allowed is False

    @patch("grantex._client.verify_grant_token")
    def test_write_scope_denies_delete_tool(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "delete_contact")

        assert result.allowed is False

    @patch("grantex._client.verify_grant_token")
    def test_write_scope_denies_admin_tool(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "manage_org")

        assert result.allowed is False

    @patch("grantex._client.verify_grant_token")
    def test_delete_scope_denies_admin_tool(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:delete",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "manage_org")

        assert result.allowed is False


# ── Token verification failures ─────────────────────────────────────────────


class TestEnforceTokenFailures:
    @patch("grantex._client.verify_grant_token")
    def test_denied_for_expired_token(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.side_effect = GrantexTokenError(  # type: ignore[attr-defined]
            "Grant token verification failed: Signature has expired"
        )
        result = client.enforce("expired.jwt.token", "salesforce", "query")

        assert result.allowed is False
        assert "Token verification failed" in result.reason
        assert result.grant_id == ""
        assert result.agent_did == ""

    @patch("grantex._client.verify_grant_token")
    def test_denied_for_invalid_token(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.side_effect = GrantexTokenError(  # type: ignore[attr-defined]
            "Grant token verification failed: invalid token"
        )
        result = client.enforce("garbage-token", "salesforce", "query")

        assert result.allowed is False
        assert "Token verification failed" in result.reason

    @patch("grantex._client.verify_grant_token")
    def test_denied_for_tampered_token(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.side_effect = GrantexTokenError(  # type: ignore[attr-defined]
            "Grant token verification failed: Signature verification failed"
        )
        result = client.enforce("tampered.jwt.token", "salesforce", "query")

        assert result.allowed is False
        assert "Token verification failed" in result.reason

    @patch("grantex._client.verify_grant_token")
    def test_denied_for_jwks_fetch_failure(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.side_effect = GrantexTokenError(  # type: ignore[attr-defined]
            "Failed to fetch JWKS"
        )
        result = client.enforce("some.jwt.token", "salesforce", "query")

        assert result.allowed is False
        assert "Token verification failed" in result.reason


# ── No matching scope for connector ─────────────────────────────────────────


class TestEnforceNoMatchingScope:
    @patch("grantex._client.verify_grant_token")
    def test_denied_when_scope_for_different_connector(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:github:admin",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "query")

        assert result.allowed is False
        assert "No scope grants access" in result.reason

    @patch("grantex._client.verify_grant_token")
    def test_denied_when_scopes_empty(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=()
        )
        result = client.enforce("fake.jwt.token", "salesforce", "query")

        assert result.allowed is False

    @patch("grantex._client.verify_grant_token")
    def test_non_tool_scopes_ignored(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("calendar:read", "payments:initiate:max_500")
        )
        result = client.enforce("fake.jwt.token", "salesforce", "query")

        assert result.allowed is False
        assert "No scope grants access" in result.reason


# ── Capped scope enforcement ────────────────────────────────────────────────


class TestEnforceCappedScopes:
    @patch("grantex._client.verify_grant_token")
    def test_amount_within_cap_allowed(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write:capped:500",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead", amount=200)

        assert result.allowed is True

    @patch("grantex._client.verify_grant_token")
    def test_amount_exceeds_cap_denied(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write:capped:500",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead", amount=750)

        assert result.allowed is False
        assert "exceeds budget cap" in result.reason
        assert "500" in result.reason

    @patch("grantex._client.verify_grant_token")
    def test_amount_exactly_at_cap_allowed(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write:capped:100",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead", amount=100)

        assert result.allowed is True

    @patch("grantex._client.verify_grant_token")
    def test_no_amount_skips_cap_check(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write:capped:100",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead")

        assert result.allowed is True

    @patch("grantex._client.verify_grant_token")
    def test_no_cap_in_scope_ignores_amount(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead", amount=999999)

        assert result.allowed is True


# ── load_manifest / load_manifests ──────────────────────────────────────────


class TestManifestLoading:
    @patch("grantex._client.verify_grant_token")
    def test_load_manifest_enables_enforcement(self, mock_verify: object) -> None:
        client = Grantex(api_key="test-key")
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )

        # Before loading, enforcement denies (no manifest)
        result = client.enforce("fake.jwt.token", "salesforce", "query")
        assert result.allowed is False
        assert "No manifest loaded" in result.reason

        # After loading, enforcement works
        client.load_manifest(_salesforce_manifest())
        result = client.enforce("fake.jwt.token", "salesforce", "query")
        assert result.allowed is True

    @patch("grantex._client.verify_grant_token")
    def test_load_manifests_loads_multiple(self, mock_verify: object) -> None:
        client = Grantex(api_key="test-key")
        client.load_manifests([_salesforce_manifest(), _github_manifest()])

        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:read",)
        )
        sf_result = client.enforce("fake.jwt.token", "salesforce", "query")
        assert sf_result.allowed is True

        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:github:write",)
        )
        gh_result = client.enforce("fake.jwt.token", "github", "create_issue")
        assert gh_result.allowed is True

    @patch("grantex._client.verify_grant_token")
    def test_load_manifest_overwrites_existing(self, mock_verify: object) -> None:
        client = Grantex(api_key="test-key")
        original = ToolManifest(
            connector="salesforce",
            tools={"query": Permission.READ},
        )
        client.load_manifest(original)

        # Now overwrite with a different manifest for the same connector
        updated = ToolManifest(
            connector="salesforce",
            tools={"export_data": Permission.ADMIN},
        )
        client.load_manifest(updated)

        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:admin",)
        )

        # Old tool no longer recognized
        result = client.enforce("fake.jwt.token", "salesforce", "query")
        assert result.allowed is False
        assert "Unknown tool" in result.reason

        # New tool recognized
        result = client.enforce("fake.jwt.token", "salesforce", "export_data")
        assert result.allowed is True


# ── Multiple scopes — best match ────────────────────────────────────────────


class TestEnforceMultipleScopes:
    @patch("grantex._client.verify_grant_token")
    def test_highest_permission_wins(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:read", "tool:salesforce:delete")
        )
        result = client.enforce("fake.jwt.token", "salesforce", "delete_contact")

        assert result.allowed is True

    @patch("grantex._client.verify_grant_token")
    def test_mixed_connector_scopes(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:github:admin", "tool:salesforce:read")
        )
        # Read-level salesforce scope should allow read tool
        result = client.enforce("fake.jwt.token", "salesforce", "query")
        assert result.allowed is True

        # But not write tool
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead")
        assert result.allowed is False


# ── agenticorg prefix scopes ────────────────────────────────────────────────


class TestEnforceAgenticorgScopes:
    @patch("grantex._client.verify_grant_token")
    def test_agenticorg_prefix_works(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("agenticorg:salesforce:write",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead")

        assert result.allowed is True

    @patch("grantex._client.verify_grant_token")
    def test_agenticorg_prefix_hierarchy(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("agenticorg:salesforce:read",)
        )
        result = client.enforce("fake.jwt.token", "salesforce", "query")
        assert result.allowed is True

        result = client.enforce("fake.jwt.token", "salesforce", "create_lead")
        assert result.allowed is False


# ── EnforceResult field population ──────────────────────────────────────────


class TestEnforceResultFields:
    @patch("grantex._client.verify_grant_token")
    def test_allowed_result_has_all_fields(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",),
            grant_id="grnt_custom",
            agent_did="did:grantex:ag_custom",
        )
        result = client.enforce("fake.jwt.token", "salesforce", "create_lead")

        assert isinstance(result, EnforceResult)
        assert result.allowed is True
        assert result.reason == ""
        assert result.grant_id == "grnt_custom"
        assert result.agent_did == "did:grantex:ag_custom"
        assert result.scopes == ["tool:salesforce:write"]
        assert result.permission == "write"
        assert result.connector == "salesforce"
        assert result.tool == "create_lead"

    @patch("grantex._client.verify_grant_token")
    def test_denied_result_preserves_context(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:read",),
            grant_id="grnt_denied",
            agent_did="did:grantex:ag_denied",
        )
        result = client.enforce("fake.jwt.token", "salesforce", "delete_contact")

        assert result.allowed is False
        assert result.grant_id == "grnt_denied"
        assert result.agent_did == "did:grantex:ag_denied"
        assert result.scopes == ["tool:salesforce:read"]
        assert result.connector == "salesforce"
        assert result.tool == "delete_contact"
        assert result.permission == "delete"


# ── wrap_tool ──────────────────────────────────────────────────────────────


class TestWrapTool:
    @patch("grantex._client.verify_grant_token")
    def test_calls_original_when_allowed(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        client.load_manifest(
            ToolManifest(
                connector="salesforce",
                tools={"create_lead": Permission.WRITE},
            )
        )

        class FakeTool:
            def _run(self, **kwargs: object) -> str:
                return "result"

        tool = FakeTool()
        client.wrap_tool(
            tool,
            connector="salesforce",
            tool_name="create_lead",
            grant_token="tok",
        )
        assert tool._run() == "result"

    @patch("grantex._client.verify_grant_token")
    def test_raises_when_denied(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:read",)
        )
        client.load_manifest(
            ToolManifest(
                connector="salesforce",
                tools={"create_lead": Permission.WRITE},
            )
        )

        class FakeTool:
            def _run(self) -> str:
                return "result"

        tool = FakeTool()
        client.wrap_tool(
            tool,
            connector="salesforce",
            tool_name="create_lead",
            grant_token="tok",
        )
        with pytest.raises(PermissionError, match="scope denied"):
            tool._run()

    @patch("grantex._client.verify_grant_token")
    def test_dynamic_token_getter(
        self, mock_verify: object, client: Grantex
    ) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        client.load_manifest(
            ToolManifest(
                connector="salesforce",
                tools={"create_lead": Permission.WRITE},
            )
        )

        class FakeTool:
            def _run(self) -> str:
                return "ok"

        tokens = ["tok-1"]
        tool = FakeTool()
        client.wrap_tool(
            tool,
            connector="salesforce",
            tool_name="create_lead",
            grant_token=lambda: tokens[0],
        )
        tool._run()
        tokens[0] = "tok-2"
        tool._run()


# ── Permissive mode ────────────────────────────────────────────────────────


class TestPermissiveMode:
    @patch("grantex._client.verify_grant_token")
    def test_permissive_allows_denied(self, mock_verify: object) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:read",)
        )
        client = Grantex(api_key="test", enforce_mode="permissive")
        client.load_manifest(
            ToolManifest(
                connector="salesforce",
                tools={"create_lead": Permission.WRITE},
            )
        )
        result = client.enforce("tok", "salesforce", "create_lead")
        assert result.allowed is True  # permissive overrides

    @patch("grantex._client.verify_grant_token")
    def test_strict_denies_normally(self, mock_verify: object) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:read",)
        )
        client = Grantex(api_key="test")
        client.load_manifest(
            ToolManifest(
                connector="salesforce",
                tools={"create_lead": Permission.WRITE},
            )
        )
        result = client.enforce("tok", "salesforce", "create_lead")
        assert result.allowed is False


# ── FastAPI GrantexEnforcer ────────────────────────────────────────────────


class TestGrantexEnforcer:
    @patch("grantex._client.verify_grant_token")
    def test_returns_result_when_allowed(self, mock_verify: object) -> None:
        mock_verify.return_value = _make_verified_grant(  # type: ignore[attr-defined]
            scopes=("tool:salesforce:write",)
        )
        client = Grantex(api_key="test")
        client.load_manifest(
            ToolManifest(
                connector="salesforce",
                tools={"create_lead": Permission.WRITE},
            )
        )
        from grantex._fastapi import GrantexEnforcer

        enforcer = GrantexEnforcer(client)
        import asyncio

        result = asyncio.get_event_loop().run_until_complete(
            enforcer(
                connector="salesforce",
                tool="create_lead",
                authorization="Bearer tok",
            )
        )
        assert result.allowed is True



"""Tests for PKCE utilities — verifier/challenge generation and validation."""
from __future__ import annotations

import base64
import hashlib
import re

import pytest

from grantex._pkce import generate_pkce, PkceChallenge


# ── Basic Generation ──────────────────────────────────────────────────────

def test_generate_pkce_returns_challenge() -> None:
    result = generate_pkce()
    assert isinstance(result, PkceChallenge)
    assert result.code_verifier
    assert result.code_challenge
    assert result.code_challenge_method == "S256"


def test_generate_pkce_returns_non_empty_strings() -> None:
    result = generate_pkce()
    assert len(result.code_verifier) > 0
    assert len(result.code_challenge) > 0
    assert len(result.code_challenge_method) > 0


# ── Verifier Validation ──────────────────────────────────────────────────

def test_verifier_is_url_safe_base64() -> None:
    result = generate_pkce()
    # Should only contain URL-safe characters (no +, /, =)
    assert re.match(r"^[A-Za-z0-9_-]+$", result.code_verifier)


def test_verifier_length_43_to_128() -> None:
    result = generate_pkce()
    assert 43 <= len(result.code_verifier) <= 128


def test_verifier_length_consistent_across_calls() -> None:
    """All verifiers should be the same length since we use 32 random bytes."""
    lengths = {len(generate_pkce().code_verifier) for _ in range(20)}
    assert len(lengths) == 1, f"Expected uniform verifier length, got {lengths}"


def test_verifier_contains_only_ascii() -> None:
    result = generate_pkce()
    assert result.code_verifier.isascii()


def test_verifier_no_padding_chars() -> None:
    """URL-safe base64 should not contain padding '=' characters."""
    for _ in range(10):
        result = generate_pkce()
        assert "=" not in result.code_verifier


# ── Challenge Validation ─────────────────────────────────────────────────

def test_challenge_is_s256_of_verifier() -> None:
    result = generate_pkce()
    # Recompute the challenge from the verifier
    digest = hashlib.sha256(result.code_verifier.encode("ascii")).digest()
    expected = base64.urlsafe_b64encode(digest).rstrip(b"=").decode("ascii")
    assert result.code_challenge == expected


def test_challenge_is_url_safe_base64() -> None:
    result = generate_pkce()
    # Should only contain URL-safe characters
    assert re.match(r"^[A-Za-z0-9_-]+$", result.code_challenge)


def test_challenge_no_padding_chars() -> None:
    for _ in range(10):
        result = generate_pkce()
        assert "=" not in result.code_challenge


def test_challenge_has_expected_length() -> None:
    """SHA-256 digest = 32 bytes, base64url without padding = 43 chars."""
    result = generate_pkce()
    assert len(result.code_challenge) == 43


def test_challenge_method_is_always_s256() -> None:
    for _ in range(10):
        result = generate_pkce()
        assert result.code_challenge_method == "S256"


# ── Uniqueness ────────────────────────────────────────────────────────────

def test_multiple_calls_produce_different_verifiers() -> None:
    a = generate_pkce()
    b = generate_pkce()
    assert a.code_verifier != b.code_verifier


def test_multiple_calls_produce_different_challenges() -> None:
    a = generate_pkce()
    b = generate_pkce()
    assert a.code_challenge != b.code_challenge


def test_100_unique_verifiers() -> None:
    verifiers = {generate_pkce().code_verifier for _ in range(100)}
    assert len(verifiers) == 100, "Expected 100 unique verifiers"


def test_100_unique_challenges() -> None:
    challenges = {generate_pkce().code_challenge for _ in range(100)}
    assert len(challenges) == 100, "Expected 100 unique challenges"


# ── Immutability ─────────────────────────────────────────────────────────

def test_pkce_challenge_is_frozen() -> None:
    result = generate_pkce()
    with pytest.raises(AttributeError):
        result.code_verifier = "mutated"  # type: ignore[misc]


def test_pkce_challenge_challenge_is_frozen() -> None:
    result = generate_pkce()
    with pytest.raises(AttributeError):
        result.code_challenge = "mutated"  # type: ignore[misc]


def test_pkce_challenge_method_is_frozen() -> None:
    result = generate_pkce()
    with pytest.raises(AttributeError):
        result.code_challenge_method = "S384"  # type: ignore[misc]


# ── Cross-verification ───────────────────────────────────────────────────

def test_verifier_challenge_pair_validates_cross_check() -> None:
    """Simulate what a server would do: verify the challenge matches."""
    pkce = generate_pkce()

    # Server receives code_verifier and code_challenge
    server_digest = hashlib.sha256(pkce.code_verifier.encode("ascii")).digest()
    server_challenge = base64.urlsafe_b64encode(server_digest).rstrip(b"=").decode("ascii")

    assert server_challenge == pkce.code_challenge


def test_wrong_verifier_does_not_match_challenge() -> None:
    """A different verifier should produce a different challenge."""
    pkce = generate_pkce()

    wrong_digest = hashlib.sha256(b"wrong-verifier").digest()
    wrong_challenge = base64.urlsafe_b64encode(wrong_digest).rstrip(b"=").decode("ascii")

    assert wrong_challenge != pkce.code_challenge


def test_deterministic_challenge_for_same_verifier() -> None:
    """Same verifier should always produce the same challenge."""
    pkce = generate_pkce()
    verifier = pkce.code_verifier

    digest = hashlib.sha256(verifier.encode("ascii")).digest()
    challenge_1 = base64.urlsafe_b64encode(digest).rstrip(b"=").decode("ascii")
    challenge_2 = base64.urlsafe_b64encode(digest).rstrip(b"=").decode("ascii")

    assert challenge_1 == challenge_2 == pkce.code_challenge



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="linkedin_ads",
    description="LinkedIn Ads API",
    tools={
        "create_campaign": Permission.WRITE,
        "get_analytics": Permission.READ,
        "create_lead_gen_form": Permission.WRITE,
        "get_targeting_criteria": Permission.READ,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="servicenow",
    description="ServiceNow ITSM API",
    tools={
        "create_incident": Permission.WRITE,
        "update_incident": Permission.WRITE,
        "submit_change_request": Permission.WRITE,
        "get_cmdb_ci": Permission.READ,
        "check_sla_status": Permission.READ,
        "get_kb_article": Permission.READ,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="google_ads",
    description="Google Ads API",
    tools={
        "search_campaigns": Permission.READ,
        "get_campaign_performance": Permission.READ,
        "mutate_campaign_budget": Permission.WRITE,
        "get_search_terms": Permission.READ,
        "create_user_list": Permission.WRITE,
    },
)



"""Tests for UsageClient — current usage and history."""
from __future__ import annotations

import json

import pytest
import respx
import httpx

from grantex import Grantex
from grantex._errors import GrantexApiError


BASE_URL = "https://api.grantex.dev"

MOCK_USAGE = {
    "developerId": "dev_01",
    "period": "2024-06-15",
    "tokenExchanges": 1500,
    "authorizations": 300,
    "verifications": 800,
    "totalRequests": 2600,
}

MOCK_USAGE_ZERO = {
    "developerId": "dev_new",
    "period": "2024-06-15",
    "tokenExchanges": 0,
    "authorizations": 0,
    "verifications": 0,
    "totalRequests": 0,
}

MOCK_HISTORY = {
    "developerId": "dev_01",
    "days": 7,
    "entries": [
        {
            "date": "2024-06-15",
            "tokenExchanges": 200,
            "authorizations": 50,
            "verifications": 100,
            "totalRequests": 350,
        },
        {
            "date": "2024-06-14",
            "tokenExchanges": 220,
            "authorizations": 55,
            "verifications": 110,
            "totalRequests": 385,
        },
        {
            "date": "2024-06-13",
            "tokenExchanges": 180,
            "authorizations": 40,
            "verifications": 90,
            "totalRequests": 310,
        },
    ],
}

MOCK_HISTORY_EMPTY = {
    "developerId": "dev_new",
    "days": 7,
    "entries": [],
}


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


# ── Current Usage ─────────────────────────────────────────────────────────

@respx.mock
def test_current_usage(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/usage").mock(
        return_value=httpx.Response(200, json=MOCK_USAGE)
    )
    result = client.usage.current()

    assert result.developer_id == "dev_01"
    assert result.period == "2024-06-15"
    assert result.token_exchanges == 1500
    assert result.authorizations == 300
    assert result.verifications == 800
    assert result.total_requests == 2600


@respx.mock
def test_current_usage_total_equals_sum(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/usage").mock(
        return_value=httpx.Response(200, json=MOCK_USAGE)
    )
    result = client.usage.current()
    assert result.total_requests == (
        result.token_exchanges + result.authorizations + result.verifications
    )


@respx.mock
def test_current_usage_zero_values(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/usage").mock(
        return_value=httpx.Response(200, json=MOCK_USAGE_ZERO)
    )
    result = client.usage.current()

    assert result.developer_id == "dev_new"
    assert result.token_exchanges == 0
    assert result.authorizations == 0
    assert result.verifications == 0
    assert result.total_requests == 0


@respx.mock
def test_current_usage_sends_get(client: Grantex) -> None:
    route = respx.get(f"{BASE_URL}/v1/usage").mock(
        return_value=httpx.Response(200, json=MOCK_USAGE)
    )
    client.usage.current()
    assert route.called
    assert route.calls[0].request.method == "GET"


@respx.mock
def test_current_usage_unauthorized(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/usage").mock(
        return_value=httpx.Response(
            401, json={"message": "Invalid API key", "code": "UNAUTHORIZED"}
        )
    )
    from grantex._errors import GrantexAuthError

    with pytest.raises(GrantexAuthError) as exc_info:
        client.usage.current()
    assert exc_info.value.status_code == 401


# ── Usage History ─────────────────────────────────────────────────────────

@respx.mock
def test_usage_history_default_days(client: Grantex) -> None:
    route = respx.get(f"{BASE_URL}/v1/usage/history?days=30").mock(
        return_value=httpx.Response(200, json=MOCK_HISTORY)
    )
    result = client.usage.history()

    assert route.called
    assert result.developer_id == "dev_01"
    assert result.days == 7  # from mock response
    assert len(result.entries) == 3


@respx.mock
def test_usage_history_custom_days(client: Grantex) -> None:
    route = respx.get(f"{BASE_URL}/v1/usage/history?days=7").mock(
        return_value=httpx.Response(200, json=MOCK_HISTORY)
    )
    result = client.usage.history(days=7)

    assert route.called
    assert result.days == 7


@respx.mock
def test_usage_history_90_days(client: Grantex) -> None:
    route = respx.get(f"{BASE_URL}/v1/usage/history?days=90").mock(
        return_value=httpx.Response(200, json={**MOCK_HISTORY, "days": 90})
    )
    result = client.usage.history(days=90)
    assert route.called
    assert result.days == 90


@respx.mock
def test_usage_history_entry_fields(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/usage/history?days=7").mock(
        return_value=httpx.Response(200, json=MOCK_HISTORY)
    )
    result = client.usage.history(days=7)
    entry = result.entries[0]

    assert entry.date == "2024-06-15"
    assert entry.token_exchanges == 200
    assert entry.authorizations == 50
    assert entry.verifications == 100
    assert entry.total_requests == 350


@respx.mock
def test_usage_history_multiple_entries(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/usage/history?days=7").mock(
        return_value=httpx.Response(200, json=MOCK_HISTORY)
    )
    result = client.usage.history(days=7)

    assert len(result.entries) == 3
    dates = [e.date for e in result.entries]
    assert dates == ["2024-06-15", "2024-06-14", "2024-06-13"]


@respx.mock
def test_usage_history_empty(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/usage/history?days=7").mock(
        return_value=httpx.Response(200, json=MOCK_HISTORY_EMPTY)
    )
    result = client.usage.history(days=7)

    assert result.developer_id == "dev_new"
    assert len(result.entries) == 0


@respx.mock
def test_usage_history_url_format(client: Grantex) -> None:
    route = respx.get(f"{BASE_URL}/v1/usage/history?days=14").mock(
        return_value=httpx.Response(200, json=MOCK_HISTORY)
    )
    client.usage.history(days=14)

    assert route.called
    request_url = str(route.calls[0].request.url)
    assert "days=14" in request_url


@respx.mock
def test_usage_history_server_error(client: Grantex) -> None:
    respx.get(f"{BASE_URL}/v1/usage/history?days=30").mock(
        return_value=httpx.Response(
            500, json={"message": "Internal server error", "code": "INTERNAL_ERROR"}
        )
    )
    with pytest.raises(GrantexApiError) as exc_info:
        client.usage.history()
    assert exc_info.value.status_code == 500



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="pagerduty",
    description="PagerDuty Incident Management API",
    tools={
        "create_incident": Permission.WRITE,
        "acknowledge_incident": Permission.WRITE,
        "resolve_incident": Permission.WRITE,
        "get_on_call": Permission.READ,
        "list_incidents": Permission.READ,
        "create_postmortem": Permission.WRITE,
    },
)



from grantex.manifest import ToolManifest, Permission

manifest = ToolManifest(
    connector="zoom",
    description="Zoom Video Communications API",
    tools={
        "create_meeting": Permission.WRITE,
        "get_recording": Permission.READ,
        "cancel_meeting": Permission.DELETE,
        "get_attendance_report": Permission.READ,
        "add_panelist": Permission.WRITE,
        "get_transcript": Permission.READ,
    },
)



"""Tests for AuditClient."""
from __future__ import annotations

import json

import pytest
import respx
import httpx

from grantex import Grantex
from grantex._types import ListAuditParams
from tests.conftest import MOCK_AUDIT_ENTRY


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


@respx.mock
def test_log_body_fields(client: Grantex) -> None:
    route = respx.post("https://api.grantex.dev/v1/audit/log").mock(
        return_value=httpx.Response(200, json=MOCK_AUDIT_ENTRY)
    )
    entry = client.audit.log(
        agent_id="ag_01HXYZ123abc",
        agent_did="did:grantex:ag_01HXYZ123abc",
        grant_id="grant_01HXYZ",
        principal_id="user_01",
        action="payment.initiated",
        metadata={"amount": 420, "currency": "USD"},
        status="success",
    )
    assert entry.entry_id == "audit_01HXYZ"
    assert entry.action == "payment.initiated"
    assert entry.metadata == {"amount": 420, "currency": "USD"}
    assert entry.status == "success"
    assert entry.prev_hash is None

    body = json.loads(route.calls[0].request.content)
    assert body["agentId"] == "ag_01HXYZ123abc"
    assert body["agentDid"] == "did:grantex:ag_01HXYZ123abc"
    assert body["grantId"] == "grant_01HXYZ"
    assert body["principalId"] == "user_01"
    assert body["action"] == "payment.initiated"
    assert body["metadata"] == {"amount": 420, "currency": "USD"}
    assert body["status"] == "success"


@respx.mock
def test_log_without_metadata(client: Grantex) -> None:
    route = respx.post("https://api.grantex.dev/v1/audit/log").mock(
        return_value=httpx.Response(200, json=MOCK_AUDIT_ENTRY)
    )
    client.audit.log(
        agent_id="ag_01HXYZ123abc",
        agent_did="did:grantex:ag_01HXYZ123abc",
        grant_id="grant_01HXYZ",
        principal_id="user_01",
        action="data.read",
    )
    body = json.loads(route.calls[0].request.content)
    assert "metadata" not in body


@respx.mock
def test_list_and_query_params(client: Grantex) -> None:
    payload = {
        "entries": [MOCK_AUDIT_ENTRY],
        "total": 1,
        "page": 1,
        "pageSize": 20,
    }
    route = respx.get("https://api.grantex.dev/v1/audit/entries").mock(
        return_value=httpx.Response(200, json=payload)
    )
    params = ListAuditParams(agent_id="ag_01HXYZ123abc", action="payment.initiated")
    response = client.audit.list(params)

    assert response.total == 1
    assert len(response.entries) == 1

    url = str(route.calls[0].request.url)
    assert "agentId=ag_01HXYZ123abc" in url
    assert "action=payment.initiated" in url


@respx.mock
def test_list_no_params(client: Grantex) -> None:
    payload = {"entries": [], "total": 0, "page": 1, "pageSize": 20}
    route = respx.get("https://api.grantex.dev/v1/audit/entries").mock(
        return_value=httpx.Response(200, json=payload)
    )
    client.audit.list()
    url = str(route.calls[0].request.url)
    assert url == "https://api.grantex.dev/v1/audit/entries"


@respx.mock
def test_get_by_id(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/v1/audit/audit_01HXYZ").mock(
        return_value=httpx.Response(200, json=MOCK_AUDIT_ENTRY)
    )
    entry = client.audit.get("audit_01HXYZ")
    assert entry.entry_id == "audit_01HXYZ"
    assert entry.prev_hash is None
    assert entry.status == "success"



from __future__ import annotations

import hashlib
import hmac

import httpx
import respx

from grantex import Grantex
from grantex._webhook import verify_webhook_signature

BASE_URL = "http://test.local"

MOCK_WEBHOOK = {
    "id": "wh_01",
    "url": "https://example.com/hooks",
    "events": ["grant.created", "grant.revoked"],
    "createdAt": "2026-02-26T00:00:00Z",
}

MOCK_WEBHOOK_WITH_SECRET = {**MOCK_WEBHOOK, "secret": "abc123secret"}


def _make_sig(payload: str | bytes, secret: str) -> str:
    raw = payload.encode() if isinstance(payload, str) else payload
    mac = hmac.new(secret.encode(), raw, hashlib.sha256)
    return "sha256=" + mac.hexdigest()


@respx.mock
def test_create_webhook() -> None:
    respx.post(f"{BASE_URL}/v1/webhooks").mock(
        return_value=httpx.Response(201, json=MOCK_WEBHOOK_WITH_SECRET)
    )
    client = Grantex(api_key="test-key", base_url=BASE_URL)
    result = client.webhooks.create(url="https://example.com/hooks", events=["grant.created"])

    assert result.id == "wh_01"
    assert result.secret == "abc123secret"
    assert "grant.created" in result.events


@respx.mock
def test_list_webhooks() -> None:
    respx.get(f"{BASE_URL}/v1/webhooks").mock(
        return_value=httpx.Response(200, json={"webhooks": [MOCK_WEBHOOK]})
    )
    client = Grantex(api_key="test-key", base_url=BASE_URL)
    result = client.webhooks.list()

    assert len(result.webhooks) == 1
    assert result.webhooks[0].id == "wh_01"
    assert result.webhooks[0].url == "https://example.com/hooks"


@respx.mock
def test_delete_webhook() -> None:
    respx.delete(f"{BASE_URL}/v1/webhooks/wh_01").mock(
        return_value=httpx.Response(204)
    )
    client = Grantex(api_key="test-key", base_url=BASE_URL)
    assert client.webhooks.delete("wh_01") is None


def test_verify_signature_valid() -> None:
    payload = '{"id":"evt_01","type":"grant.created","data":{}}'
    secret = "my-webhook-secret"
    sig = _make_sig(payload, secret)
    assert verify_webhook_signature(payload, sig, secret) is True


def test_verify_signature_invalid() -> None:
    assert verify_webhook_signature("payload", "sha256=badsig", "secret") is False


def test_verify_signature_wrong_secret() -> None:
    payload = '{"id":"evt_01"}'
    sig = _make_sig(payload, "correct-secret")
    assert verify_webhook_signature(payload, sig, "wrong-secret") is False


def test_verify_signature_bytes_payload() -> None:
    payload = b'{"id":"evt_01","type":"grant.created"}'
    secret = "my-secret"
    sig = _make_sig(payload, secret)
    assert verify_webhook_signature(payload, sig, secret) is True
